/*
Copyright (c) 2010, Dirk Krause
All rights reserved.

Redistribution and use in source and binary forms,
with or without modification, are permitted provided
that the following conditions are met:

* Redistributions of source code must retain the above
  copyright notice, this list of conditions and the
  following disclaimer.
* Redistributions in binary form must reproduce the above
  opyright notice, this list of conditions and the following
  disclaimer in the documentation and/or other materials
  provided with the distribution.
* Neither the name of the Dirk Krause nor the names of
  contributors may be used to endorse or promote
  products derived from this software without specific
  prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED.
IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
*/

package dirk_krause.tools;

import java.net.URL;
import java.awt.Rectangle;
import java.awt.Dimension;
import java.awt.Toolkit;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;
import java.awt.print.Printable;

import javax.swing.JFrame;
import javax.swing.JOptionPane;

/**
 * Application main windows class.
 * After constructing your window use pack(), choosePosition()
 * and setVisible(true).
 * @author krause
 */
public abstract class DkFrame
        extends JFrame
        implements WindowListener
{
    /** Number of instances. */
    private static int numberOfInstances = 0;

    /** General options tool. */
    protected DkTool dkt = null;

    /** GUI options tool. */
    protected DkGuiTool dkgt = null;

    /** Border (space) between GUI elements. */
    protected int border = 5;

    /** Flag: Can use System.exit() when closing final window. */
    protected boolean exitOnClose = true;

    /** Screen width. */
    protected int scw = -1;

    /** Screen height. */
    protected int sch = -1;

    /** x, y, width, height */
    protected int[] xy = { -1, -1, -1, -1 };

    protected static String[] pkInt = {
        "ui.window.x", "ui.window.y", "ui.window.w", "ui.window.h"
    };

    /** Message dialog. */
    protected DkMsgDialog messageDialog = null;

    /** Dialog box for GUI settings. */
    protected DkGuiDialog guiDialog = null;

    /** Print client. */
    protected DkPrintClient prcl = null;

    /**
     * Constructor.
     * @param a Application name.
     * @param d General tool.
     * @param g GUI settings tool.
     */
    public DkFrame(String a, DkTool d, DkGuiTool g) {
        super(a);
        dkt = d;
        dkgt = g;
        synchronized(this.getClass()) { numberOfInstances++; }
        addWindowListener(this);
        border = g.getComponentBorder();
        try {
            Dimension dim = Toolkit.getDefaultToolkit().getScreenSize();
            scw = dim.width; sch = dim.height;
        } catch(Exception e) { scw = 800; sch = 600; }
    }


    /**
     * Choose window position (either centered on the screen
     * or the last position saved when using the program before).
     */
    public void choosePosition() {
        dkt.getMultiplePreferenceInts(pkInt, xy);
        Rectangle rv = getBounds();
        // Curent sizes larger than saved ?
        if(rv.width > xy[2]) xy[2] = rv.width;
        if(rv.height > xy[3]) xy[3] = rv.height;
        // Saved sizes larger than current ?
        if((xy[2] > rv.width) || (xy[3] > rv.height)) {
            setSize(xy[2], xy[3]);
        }
        // Set location
        if(xy[0] < 0) { xy[0] = (scw - xy[2]) / 2; }
        if(xy[1] < 0) { xy[1] = (sch - xy[3]) / 2; }
        // Correct location
        if(xy[0] + xy[2] >= scw) { xy[0] = scw - xy[2]; }
        if(xy[1] + xy[3] >= sch) { xy[1] = sch - xy[3]; }
        if(xy[0] < 0) xy[0] = 0;
        if(xy[1] < 0) xy[1] = 1;
        // Establish location
        setLocation(xy[0], xy[1]);
    }


    /**
     * Decrement the number of instances.
     * @return
     */
    private int decreaseNumberOfInstances() {
        int back = 0;
        synchronized(this.getClass()) {
            numberOfInstances--;
            if(numberOfInstances < 0) {
                numberOfInstances = 0;
            }
            back = numberOfInstances;
        }
        return back;
    }


    /**
     * Show a message dialog.
     * @param t Title.
     * @param m Message text.
     * @param h Flag: Text is HTML.
     */
    public void showMessageDialog(String t, String m, boolean h) {
        DkMsgDialog d;
        synchronized(this) { d = messageDialog; }
        if(d == null) {
            d = new DkMsgDialog(this, dkt, dkgt);
            synchronized(this) { messageDialog = d; }
            DkTool.chooseModalPosition(scw, sch, this, d);
        }
        d.setTitle(t);
        d.setHtml(h);
        d.setEditorText(m);
        d.setModal(true); d.setVisible(true);
        d.repaint();
    }


    /**
     * Show reource from an URL in the message dialog.
     * @param t Title.
     * @param u URL.
     * @param h Flag: URL contains HTML.
     */
    public void showMessageDialogForUrl(String t, URL u, boolean h) {
        DkMsgDialog d;
        synchronized(this) { d = messageDialog; }
        if(d == null) {
            d = new DkMsgDialog(this, dkt, dkgt);
            synchronized(this) { messageDialog = d; }
            DkTool.chooseModalPosition(scw, sch, this, d);
        }
        d.setTitle(t);
        d.setHtml(h);
        d.setEditorPage(u);
        d.setModal(true); d.setVisible(true);
        d.repaint();
    }


    /**
     * Internal function to show a message box.
     * @param t Title.
     * @param m Message text.
     * @param l Log level (JOptionPane.xxx_MESSAGE).
     * @param h Flag: Text is HTML.
     */
    private void internalMessageBox(String t, String m, int l, boolean h) {
        boolean failed = false;
        try {
            JOptionPane.showMessageDialog(this, m, t, l);
        } catch(Exception e) { failed = true; }
        if(failed) {
            showMessageDialog(t, m, h);
        }
    }


    /**
     * Show an information message box.
     * @param t Title.
     * @param m Message text
     * @param h Flag: Text is HTML.
     */
    public void msgInfo(String t, String m, boolean h) {
        internalMessageBox(t, m, JOptionPane.INFORMATION_MESSAGE, h);
    }


    /**
     * Show an warning message box.
     * @param t Title.
     * @param m Message text
     * @param h Flag: Text is HTML.
     */
    public void msgWarning(String t, String m, boolean h) {
        internalMessageBox(t, m, JOptionPane.WARNING_MESSAGE, h);
    }


    /**
     * Show an error message box.
     * @param t Title.
     * @param m Message text
     * @param h Flag: Text is HTML.
     */
    public void msgError(String t, String m, boolean h) {
        internalMessageBox(t, m, JOptionPane.ERROR_MESSAGE, h);
    }


    /**
     * Show GUI setup dialog.
     */
    public void showGuiDialog() {
        DkGuiDialog d;
        synchronized(this) { d = guiDialog; }
        if(d == null) {
            d = new DkGuiDialog(this, dkt, dkgt);
            synchronized(this) { guiDialog = d; }
            DkTool.chooseModalPosition(scw, sch, this, d);
        }
        d.dataGet();
        d.setModal(true); d.setVisible(true);
        d.repaint();
    }


    /**
     * Get the print client object for this window, create it if
     * necessary.
     * @return  The print client object.
     */
    private DkPrintClient getPrintClient() {
        DkPrintClient back = null;
        synchronized(this) { back = prcl; }
        if(prcl == null) {
            back = new DkPrintClient(dkt);
            synchronized(this) { prcl = back; }
        }
        return back;
    }


    /**
     * Send print job to printer.
     * @param psData    PS print data (may be null).
     * @param pr        Printable object.
     * @param vs        Flag: Verbose on success.
     * @param ve        Flag: Verbose on errors.
     */
    public
    void sendToPrinter(String psData, Printable pr, boolean vs, boolean ve) {
        DkPrintClient pc = getPrintClient();
        DkPrintResult pres = pc.handleJob(this, psData, pr);
        if(pres.getResult()) {
            if(vs) {
                msgInfo(pres.getTitle(), pres.getText(), false);
            }
        } else {
            if(ve) {
                msgError(pres.getTitle(), pres.getText(), false);
            }
        }
    }


    /**
     * Send print job to printer.
     * @param psData    PS print data (may be null).
     * @param pr        Printable object.
     */
    public void sendToPrinter(String psData, Printable pr) {
        sendToPrinter(psData, pr, true, true);
    }


    /**
     * Send print job to printer.
     * @param psData    PS print data (may be null).
     * @param vs        Flag: Verbose on success.
     * @param ve        Flag: Verbose on errors.
     */
    public void sendToPrinter(String psData, boolean vs, boolean ve) {
        sendToPrinter(psData, null, vs, ve);
    }


    /**
     * Send print job to printer.
     * @param psData    PS print data.
     */
    public void sendToPrinter(String psData) {
        sendToPrinter(psData, null, true, true);
    }


    /**
     * Send print job to printer.
     * @param pr        Printable object.
     * @param vs        Flag: Verbose on success.
     * @param ve        Flag: Verbose on errors.
     */
    public void sendToPrinter(Printable pr, boolean vs, boolean ve) {
        sendToPrinter(null, pr, vs, ve);
    }


    /**
     * Send print job to printer.
     * @param pr        Printable object.
     */
    public void sendToPrinter(Printable pr) {
        sendToPrinter(null, pr, true, true);
    }


    /**
     * Cleanup method, invoked when the window is closed.
     *
     * @param isFinalWindow
     * Indicates whether or not this was the last visible
     * window.
     */
    abstract protected void cleanup(boolean isFinalWindow);

    protected void setExitOnClose(boolean b) {
        synchronized(this) { exitOnClose = b; }
    }

    /**
     * Callback method, invoked when a window is closed.
     *
     * @param e
     * The window event indicating that the window is closed.
     */
    @Override
    public void windowClosing(WindowEvent e) {
        int now = decreaseNumberOfInstances();
        removeWindowListener(this);
        setVisible(false);
        if(now == 0) {
            Rectangle rv = new Rectangle();
            getBounds(rv);
            xy[0] = rv.x; xy[1] = rv.y;
            xy[2] = rv.width; xy[3] = rv.height;
	    dkt.setMultiplePreferenceInts(pkInt, xy);
            dkt.flushPreferences();
        }
        cleanup((now == 0) ? true : false);
        removeAll();
        dkt = null;
        dkgt = null;
        dispose();
        if(now == 0) {
            System.gc();
            System.runFinalization();
            boolean canExit;
            synchronized(this) { canExit = exitOnClose; }
            if(canExit) {
                try { System.exit(0); } catch(Exception exc) {}
            }
        }
    }

    /** Unused window event handler. */
    @Override
    public void windowActivated(WindowEvent e) {}

    /** Unused window event handler. */
    @Override
    public void windowClosed(WindowEvent e) {}

    /** Unused window event handler. */
    @Override
    public void windowDeactivated(WindowEvent e) {}

    /** Unused window event handler. */
    @Override
    public void windowDeiconified(WindowEvent e) {}

    /** Unused window event handler. */
    @Override
    public void windowIconified(WindowEvent e) {}

    /** Unused window event handler. */
    @Override
    public void windowOpened(WindowEvent e) {}

}
