/*
Copyright (c) 2010, Dirk Krause
All rights reserved.

Redistribution and use in source and binary forms,
with or without modification, are permitted provided
that the following conditions are met:

* Redistributions of source code must retain the above
  copyright notice, this list of conditions and the
  following disclaimer.
* Redistributions in binary form must reproduce the above
  opyright notice, this list of conditions and the following
  disclaimer in the documentation and/or other materials
  provided with the distribution.
* Neither the name of the Dirk Krause nor the names of
  contributors may be used to endorse or promote
  products derived from this software without specific
  prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED.
IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
*/

package dirk_krause.tools;

import java.util.Enumeration;
import java.util.HashSet;
import java.awt.Font;
import java.awt.Toolkit;
import javax.swing.UIManager;
import javax.swing.LookAndFeel;
import javax.swing.UIDefaults;
import javax.swing.plaf.FontUIResource;

/**
 * GUI configuration tool.
 * The constructor of the first instance changes the GUI configuration.
 * Settings are read from the DkTool instance provided to the constructor.
 * First we choose a look-and-feel, either system (default) or
 * cross-platform look-and-feel.
 * Next we inspect all fonts and font resources. If there are font sizes
 * in use less than a specified minimum we correct the font size.
 * Sometimes we want to replace all fonts of a family (i.e. Dialog) by
 * another family (i.e. Verdana). These replacements are also done here.
 * @author krause
 */
public class DkGuiTool
{
    /** General tool. */
    private DkTool dkt = null;

    /** First instance, must initialize. */
    private static boolean is_first = true;

    /** Integer values. */
    private int[] myInt = {
        5,  //  0: Component border
        0,  //  1: Minimum font size
        0,  //  2: LAF index
    };

    /** Keys to store integer values. */
    private String[] myPkInt = {
        "ui.component-border",
        "ui.font-size.minimum",
        "ui.look-and-feel",
    };

    /** String values. */
    private String[] myStr = {
        null,   //  0: Old font (must be replaced).
        null,   //  1: New font (Replacement font).
    };

    /** Keys to store string values. */
    private String[] myPkStr = {
        "ui.font-replacement.old",
        "ui.font-replacement.new",
    };

    /** Boolean values. */
    private boolean[] myBool = {
        false,  // 0: Replace fonts
    };

    /** Keys to store boolean values. */
    private String[] myPkBool = {
        "ui.font-replacement.enabled",
    };

    /** Used fonts (name+size) before and after replacement. */
    private String[] fontReplacementInfo = {
        "", ""
    };

    /**
     * Constructor.
     * @param d          General tool.
     * @param useToolkit Flag: Allowed to find minimum font size automatically.
     */
    public DkGuiTool(DkTool d, boolean useToolkit) {
        dkt = d;
        boolean mustInitialize = false;
        HashSet<String> s1 = new HashSet<String>();
        HashSet<String> s2 = new HashSet<String>();
        StringBuilder sb1 = new StringBuilder();
        StringBuilder sb2 = new StringBuilder();
        if(useToolkit) {
            try {
                int scres = Toolkit.getDefaultToolkit().getScreenResolution();
                myInt[1] = (int)Math.round((14.0 * (double)scres) / 98.0);
            } catch(Exception e) { myInt[1] = 0; }
        }
        dkt.getMultiplePreferenceInts(myPkInt, myInt);
        dkt.getMultiplePreferenceStrings(myPkStr, myStr);
        dkt.getMultiplePreferenceBools(myPkBool, myBool);
        synchronized(this.getClass()) { mustInitialize = is_first; }
        if(mustInitialize) {
            synchronized(this.getClass()) { is_first = false; }
            String s = null;
            switch(myInt[2]) {
                case 0: {   // System look and feel
                    try {
                        s = UIManager.getSystemLookAndFeelClassName();
                    } catch(Exception e) { s = null; }
                } break;
                case 1: {   // Java look and feel
                    try {
                        s = UIManager.getCrossPlatformLookAndFeelClassName();
                    } catch(Exception e) { s = null; }
                } break;
            }
            if(s != null) {
                UIDefaults[] uids = new UIDefaults[2];
                uids[0] = null; uids[1] = null;
                // Set choosen LAF
                try {
                    UIManager.setLookAndFeel(s);
                } catch(Exception e) {}
                // Correct font usage
                try {
                    LookAndFeel laf = UIManager.getLookAndFeel();
                    uids[0] = laf.getDefaults();
                    uids[1] = UIManager.getDefaults();
                    for(UIDefaults u: uids) {
                        if(u != null) {
                            Object k;   /* Key object. */
                            Object v;   /* Value object. */
                            String fna; /* Font name. */
                            int fst;    /* Font style. */
                            int fsi;    /* Font size. */
                            boolean mr; /* Flag: Must replace. */
                            Font f;     /* Font */
                            FontUIResource fuir;    /* Font resource */
                            HashSet<Object> hs = new HashSet<Object>();
                            Enumeration<Object> en = u.keys();
                            boolean cc = true;
                            while((cc) && en.hasMoreElements()) {
                                cc = false; k = null; v = null;
                                try {
                                    k = en.nextElement();
                                    v = u.get(k);
				    if(k != null) { cc = true; }
                                } catch(Exception e) { cc = false; }
                                if(( k!= null) && (v != null)) {
                                    hs.add(k);
                                }
                            }
                            Object[] ka;
                            ka = hs.toArray();
                            for(Object o: ka) {
                                mr = false;
                                try {
                                    v = u.get(o);
                                    if(v != null) {
                                        if(v instanceof Font) {
                                            f = (Font)v;
                                            fna = f.getName();
                                            fst = f.getStyle();
                                            fsi = f.getSize();
                                            s1.add(fna + " " + fsi);
                                            if(myBool[0]) {
                                                if(myStr[0] != null) {
                                                    if(myStr[0].equals(fna)) {
                                                        if(myStr[1] != null) {
                                                            fna = myStr[1];
                                                            mr = true;
                                                        }
                                                    }
                                                }
                                            }
                                            if(myInt[1] > 0) {
                                                if(fsi < myInt[1]) {
                                                    fsi = myInt[1];
                                                    mr = true;
                                                }
                                            }
                                            if(mr) {
                                                u.put(o, new Font(fna,fst,fsi));
                                            }
					    s2.add(fna + " " + fsi);
                                        } else {
                                            if(v instanceof FontUIResource) {
                                                fuir = (FontUIResource)v;
                                                fna = fuir.getName();
                                                fst = fuir.getStyle();
                                                fsi = fuir.getSize();
                                                s1.add(fna + " " + fsi);
                                                if(myBool[0]) {
                                                    if(myStr[0] != null) {
                                                        if(myStr[0].equals(fna))
                                                        {
                                                            if(myStr[1] != null)
                                                            {
                                                                fna = myStr[1];
                                                                mr = true;
                                                            }
                                                        }
                                                    }
                                                }
                                                if(myInt[1] > 0) {
                                                    if(fsi < myInt[1]) {
                                                        fsi = myInt[1];
                                                        mr = true;
                                                    }
                                                }
                                                if(mr) {
                                                    u.put(
                                                        o,
                                                        new
                                                        FontUIResource(
                                                            fna,fst,fsi
                                                        )
                                                    );
                                                }
						s2.add(fna + " " + fsi);
                                            }
                                        }
                                    }
                                } catch(Exception e) {}
                            }

                        }
                    }
                    boolean isNotFirst = false;
                    for(Object o1: s1) {
                        if(isNotFirst) {
                            sb1.append(", ");
                        } isNotFirst = true;
                        sb1.append((String)o1);
                    }
                    isNotFirst = false;
                    for(Object o1: s2) {
                        if(isNotFirst) {
                            sb2.append(", ");
                        } isNotFirst = true;
                        sb2.append((String)o1);
                    }
                } catch(Exception e) { uids[0] = null; uids[1] = null; }
            }
        }
        fontReplacementInfo[0] = new String(sb1);
        fontReplacementInfo[1] = new String(sb2);
    }



    /**
     * Convenience constructor.
     * @param d General tool.
     */
    public DkGuiTool(DkTool d) {
        this(d, true);
    }


    /**
     * Get space between components.
     * @return  Space between components in pixels.
     */
    public int getComponentBorder() {
        int back;
        synchronized(this) { back = myInt[0]; }
        return back;
    }


    /**
     * Set space between components.
     * @param i New space between components in pixels.
     */
    public void setComponentBorder(int i) {
        int j = i;
        if(j < 0) { j = 0; }
        if(j > 20) { j = 20; }
        synchronized(this) { myInt[0] = j; }
    }


    /**
     * Get index of current look-and-feel
     * (0=system LAF, 1=cross-platform LAF).
     * @return  Index of current look-and-feel
     */
    public int getLafIndex() {
        int back;
        synchronized(this) { back = myInt[2]; }
        return back;
    }


    /**
     * Set index of look-and-feel.
     * @param i Index of new look-and-feel.
     */
    public void setLafIndex(int i) {
        int j = i;
        if(j < 0) { j = 0; }
        if(j > 2) { j = 2; }
        synchronized(this) { myInt[2] = j; }
    }


    /**
     * Get minimum font size.
     * @return  The minimum font size (0 for no minimum size required).
     */
    public int getMinimumFontSize() {
        int back;
        synchronized(this) { back = myInt[1]; }
        return back;
    }


    /**
     * Set minimum font size (0 for no minimum size required).
     * @param i New minimum font size.
     */
    public void setMinimumFontSize(int i) {
        int j = i;
        if(j < 0) { j = 0; }
        synchronized(this) { myInt[1] = j; }
    }


    /**
     * Set font family name to replace.
     * @param s Name of font family to replace.
     */
    public void setOldFont(String s) {
        String t;
        if(s != null) { t = new String(s); }
        else { t = s; }
        synchronized(this) { myStr[0] = t; }
    }


    /**
     * Set replacement font family name.
     * @param s The replacement font family name.
     */
    public void setNewFont(String s) {
        String t;
        if(s != null) { t = new String(s); }
        else { t = s; }
        synchronized(this) { myStr[1] = t; }
    }


    /**
     * Get font family name to replace.
     * @return  The font family name to replace.
     */
    public String getOldFont() {
        String back;
        synchronized(this) { back = myStr[0]; }
        if(back != null) { back = new String(back); }
        return back;
    }


    /**
     * Get replacement font family name.
     * @return  The replacement font family name.
     */
    public String getNewFont() {
        String back;
        synchronized(this) { back = myStr[1]; }
        if(back != null) { back = new String(back); }
        return back;
    }


    /**
     * Get the replace-fonts flag.
     * @return  Replace fonts setting.
     */
    public boolean getReplaceFonts() {
        boolean back;
        synchronized(this) { back = myBool[0]; }
        return back;
    }


    /**
     * Enable/disable font replacement.
     * @param b New replacement flag value.
     */
    public void setReplaceFonts(boolean b) {
        synchronized(this) { myBool[0] = b; }
    }


    /**
     * Get list of fonts (family+size) used before replacement.
     * @return String containing the list.
     */
    public String getFontsBeforeReplacement() {
        String back;
        synchronized(this) { back = fontReplacementInfo[0]; }
        if(back != null) { back = new String(back); }
        return back;
    }


    /**
     * Get list of fonts (family+size) used after replacement.
     * @return String containing the list.
     */
    public String getFontsAfterReplacement() {
        String back;
        synchronized(this) { back = fontReplacementInfo[1]; }
        if(back != null) { back = new String(back); }
        return back;
    }



    /**
     * Save settings (i.e. when closing the DkGuiDialog).
     * @param stage 0=application, 1=+group, 2=+all.
     */
    public void save(int stage) {
        boolean g = false;
        boolean a = false;
        if(stage > 0) {
            g = true;
            if(stage > 1) {
                a = true;
            }
        }
        int[] i;
        String[] s;
        boolean[] b;
        int j;
        synchronized(this) {
            i = new int[myInt.length];
            for(j = 0; j < myInt.length; j++) i[j] = myInt[j];
            s = new String[myStr.length];
            for(j = 0; j < myStr.length; j++) s[j] = myStr[j];
            b = new boolean[myBool.length];
            for(j = 0; j < myBool.length; j++) b[j] = myBool[j];
        }
        dkt.setMultiplePreferenceBools(myPkBool, b, g, a);
        dkt.setMultiplePreferenceInts(myPkInt, i, g, a);
        dkt.setMultiplePreferenceStrings(myPkStr, s, g, a);
        dkt.flushPreferences();
    }
    
}
