/*
Copyright (c) 2010, Dirk Krause
All rights reserved.

Redistribution and use in source and binary forms,
with or without modification, are permitted provided
that the following conditions are met:

* Redistributions of source code must retain the above
  copyright notice, this list of conditions and the
  following disclaimer.
* Redistributions in binary form must reproduce the above
  opyright notice, this list of conditions and the following
  disclaimer in the documentation and/or other materials
  provided with the distribution.
* Neither the name of the Dirk Krause nor the names of
  contributors may be used to endorse or promote
  products derived from this software without specific
  prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED.
IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
*/

package dirk_krause.tools;

import java.awt.print.Printable;
import java.awt.print.PageFormat;
import java.awt.print.PrinterJob;

import javax.swing.JFrame;
import javax.swing.JOptionPane;

import javax.print.Doc;
import javax.print.SimpleDoc;
import javax.print.DocFlavor;
import javax.print.PrintService;
import javax.print.PrintServiceLookup;
import javax.print.DocPrintJob;
import javax.print.attribute.PrintRequestAttributeSet;
import javax.print.attribute.HashPrintRequestAttributeSet;


/**
 * Objects of this class can be used to send print jobs to the print
 * system. The job can be either a Printable or a
 * String or a byte array containing
 * PostScript.
 * The webstart variant can handle Printable objects only.
 * @author krause
 */
public class DkPrintClient {

    /** Default strings to show if not string file is found. */
    public static final String[] def_t = {
        /*   0 */ "Success",
        /*   1 */ "Error",
        /*   2 */ "Your print job was successfully submitted to queue \"",
        /*   3 */ "\".",
        /*   4 */ "Your print job was successfully submitted to the print system.",
        /*   5 */ "Failed to submit print job to print system.",
        /*   6 */ "Print queue",
        /*   7 */ "Please choose a printer:",
        /*   8 */ "OK",
        /*   9 */ "Cancel",
        /*  10 */ "There was no print service found.",
    };

    /** Strings to show, either from file or the default strings. */
    public static String[] t = def_t;

    /** General tool. */
    protected DkTool dkt = null;

    /** Document flavor for PS documents. */
    public static DocFlavor dfPs = DocFlavor.BYTE_ARRAY.POSTSCRIPT;

    /** Document flavor for Printable documents. */
    public static DocFlavor dfPr = DocFlavor.SERVICE_FORMATTED.PRINTABLE;

    /** Attribute set to print PS contents. */
    protected PrintRequestAttributeSet asetPs = null;

    /** Attribute set to print a Printable object. */
    protected PrintRequestAttributeSet asetPr = null;

    /** Print services which can handle PS. */
    protected PrintService[] psPs = null;

    /** Print services which can handle a Printable. */
    protected PrintService[] psPr = null;

    /** Names of the print services for PS. */
    protected String[] psNamesPs = null;

    /** Names of the print services for a Printable. */
    protected String[] psNamesPr = null;

    /** Currently selected PS printer. */
    protected int psSelPs = -1;

    /** Currently selected Printable printer. */
    protected int psSelPr = -1;

    /**
     * Constructor.
     * @param d General tool.
     */
    public DkPrintClient(DkTool d) {
        dkt = d;
        String[] new_t = d.getLocalizedArray(this, "dkprint.txt", "UTF-8");
        if(new_t != null) {
            if(new_t.length >= def_t.length) {
                t = new_t;
            }
        }
        // Create print attribute sets.
        asetPs = new HashPrintRequestAttributeSet();
        asetPr = new HashPrintRequestAttributeSet();
        // Look up print services
        psPs = PrintServiceLookup.lookupPrintServices(dfPs, asetPs);
        psPr = PrintServiceLookup.lookupPrintServices(dfPr, asetPr);
        if(psPs != null) {
            if(psPs.length > 0) {
                psNamesPs = new String[psPs.length];
                for(int i = 0; i < psPs.length; i++) {
                    psNamesPs[i] = new String(psPs[i].getName());
                }
            }
        }
        if(psPr != null) {
            if(psPr.length > 0) {
                psNamesPr = new String[psPr.length];
                for(int i = 0; i < psPr.length; i++) {
                    psNamesPr[i] = new String(psPr[i].getName());
                }
            }
        }
        PrintService psDef = PrintServiceLookup.lookupDefaultPrintService();
        if(psDef != null) {
            if(psNamesPs != null) {
                psSelPs = DkTool.getStringArrayIndex(psNamesPs,psDef.getName());
            }
            if(psNamesPr != null) {
                psSelPr = DkTool.getStringArrayIndex(psNamesPr,psDef.getName());
            }
        }
    }

    private byte charToByte(char c) {
        byte back;
	int x = c;
	while(x >  127) { x = x - 256; }
	while(x < -128) { x = x + 256; }
	back = (byte)x;
	return back;
    }

    /**
     * Submit one print job.
     * @param w         The window originating the request.
     * @param psData    PS data to print.
     * @param pr        Printable to print.
     * @return          Result object to indicate success/error and detailed
     * information.
     */
    public DkPrintResult handleJob(JFrame w, String psData, Printable pr)
    {
        DkPrintResult back = null;
        boolean done = false;
        boolean aborted = false;
        int i;
        int psi;
        if(psData != null) {
            if(psNamesPs != null) {
                synchronized(this) { psi = psSelPs; }
                if(psi < 0) { psi = 0; }
                String psName = (String)JOptionPane.showInputDialog(
                    w, t[7], t[6], JOptionPane.PLAIN_MESSAGE, null,
                    psNamesPs, psNamesPs[psi]
                );
                if(psName != null) {
                    if(psName.length() > 0) {
                        psi = DkTool.getStringArrayIndex(psNamesPs, psName);
                        if(psi > -1) {
                            synchronized(this) { psSelPs = psi; }
                            int psl = psData.length();
                            if(psl > 0) {
                                byte[] ba;
                                if(psData.charAt(psl - 1) != 0x04) {
                                    ba = new byte[psl+1];
                                    for(i = 0; i < psl; i++) {
                                        ba[i] = charToByte(psData.charAt(i));
                                    }
                                    ba[i] = 0x04;
                                } else {
                                    ba = new byte[psl];
                                    for(i = 0; i < psl; i++) {
                                        ba[i] = charToByte(psData.charAt(i));
                                    }
                                }
                                try {
                                    DocPrintJob pj = psPs[psi].createPrintJob();
                                    Doc doc = new SimpleDoc(ba, dfPs, null);
                                    pj.print(doc, asetPs);
                                    done = true;
                                    back = new DkPrintResult(
                                        true, t[0],
                                        t[2] + psNamesPs[psi] + t[3]
                                    );
                                }
                                catch(Exception e) {
                                    done = false; back = null;
                                }
                            }
                        }
                    }
                }
            }
        }
        if(!done) {
            if(pr != null) {
                try {
                    PrinterJob pj = PrinterJob.getPrinterJob();
                    if(pj.printDialog()) {
                        pj.setPrintable(pr);
                        pj.print();
                        done = true;
                        back = new DkPrintResult(
                            true, t[0], t[4]
                        );
                    } else { aborted = true; }
                }
                catch(Exception e) {
                    done = false; back = null;
                }
            }
            if((!done) && (!aborted)) {
 		if(psNamesPr != null) {
		    synchronized(this) { psi = psSelPr; }
		    if(psi < 0) { psi = 0; }
		    String prName = (String)JOptionPane.showInputDialog(
		        w, t[7], t[6], JOptionPane.PLAIN_MESSAGE, null,
			psNamesPr, psNamesPr[psi]
		    );
		    if(prName != null) {
		        if(prName.length() > 0) {
			    psi = DkTool.getStringArrayIndex(psNamesPr,prName);
			    if(psi > 0) {
			        try {
				    DocPrintJob pj = psPr[psi].createPrintJob();
				    Doc doc = new SimpleDoc(pr, dfPr, null);
				    pj.print(doc, asetPr);
				    done = true;
				    back = new DkPrintResult(
				        true, t[0],
					t[2] + psNamesPr[psi] + t[3]
				    );
				}
				catch(Exception e) {
				    done = false; back = null;
				}
			    }
			}
		    }
		}
            }
        }
        if(back == null) {
            back = new DkPrintResult(false, t[1], t[5]);
        }
        return back;
    }
}


