/*
Copyright (c) 2010, Dirk Krause
All rights reserved.

Redistribution and use in source and binary forms,
with or without modification, are permitted provided
that the following conditions are met:

* Redistributions of source code must retain the above
  copyright notice, this list of conditions and the
  following disclaimer.
* Redistributions in binary form must reproduce the above
  opyright notice, this list of conditions and the following
  disclaimer in the documentation and/or other materials
  provided with the distribution.
* Neither the name of the Dirk Krause nor the names of
  contributors may be used to endorse or promote
  products derived from this software without specific
  prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED.
IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
*/

package dirk_krause.wstools;

import java.util.prefs.Preferences;
import java.util.ArrayList;
import java.util.StringTokenizer;
import java.util.Properties;
import java.util.Enumeration;
import java.util.Collections;
import java.util.TreeMap;
import java.util.Iterator;
import java.util.Date;
import java.util.TimeZone;
import java.util.Calendar;
import java.util.GregorianCalendar;

import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.BufferedReader;

import java.awt.Font;
import java.awt.Color;
import java.awt.Rectangle;
import java.awt.Component;
import java.awt.Image;
import java.awt.image.BufferedImage;

import java.net.URL;
import java.net.InetAddress;
import java.net.NetworkInterface;

import javax.swing.ImageIcon;

import javax.imageio.ImageIO;

/**
 *
 * @author krause
 */
public class DkTool
{
    /** Run-type: Application. */
    public static final int RUN_TYPE_APPLICATION = 0;

    /** Run-type: Web-start application. */
    public static final int RUN_TYPE_WEBSTART = 1;

    /** Run-type: Applet. */
    public static final int RUN_TYPE_APPLET = 2;

    /** Static texts used by this class. */
    public static final String[] stt = {
        /*  0 */ "\n",
        /*  1 */ " ",
        /*  2 */ "all",
        /*  3 */ "user.name",
        /*  4 */ "user.language",
        /*  5 */ "user.home",
        /*  6 */ "user.dir",
        /*  7 */ "/",
        /*  8 */ "",
        /*  9 */ "http://",
        /* 10 */ "file://",
        /* 11 */ ":",
        /* 12 */ "file.separator",
        /* 13 */ "path.separator",
        /* 14 */ "\\",
        /* 15 */ "HOSTNAME",
        /* 16 */ "COMPUTERNAME",
        /* 17 */ "config.txt",
        /* 18 */ "[",
        /* 19 */ "*",
        /* 20 */ "#",
        /* 21 */ "resources.directory.fonts",
        /* 22 */ "resources.directory.images",
        /* 23 */ "resources.directory.application",
	/* 24 */ ".ttf",
	/* 25 */ ".TTF",
    };

    /** Hexadecimal characters, used by encodeToHtml. */
    protected static final char[] hexCharacters = {
        '0', '1', '2', '3', '4', '5', '6', '7',
        '8', '9', 'A', 'B', 'C', 'D', 'E', 'F'
    };

    /** Strings which can be converted into boolean values. */
    protected static final String[] theBooleanStrings = {
        "y", "yes", "on", "true", "ok", "1", "+",
        "n", "no", "off", "false", "0", "-"
    };
    
    /** Index of first "no" entry in theBooleanStrings. */
    protected static final int first_no_boolean = 7;

    /** Allowed character set in PS files. */
    protected static final String psAllowedCharacters =
    "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789+-*/?.,;";

    /** Application name. */
    private String appName = null;

    /** Application group name. */
    private String grpName = null;

    /** Run type of object. */
    private int runType = RUN_TYPE_WEBSTART;

    /** Preferences by user. */
    private Preferences[] myprefs = null;

    /** Properties (system-wide settings, read-only). */
    private Properties props = null;

    /** My properties, self-set. */
    private Properties mypro = null;

    /** Resource directories. */
    private ArrayList<String> resourceBaseList = null;

    /** User name of the current user. */
    private String userName = null;

    /** Language setting (without region). */
    private String userLanguage = null;

    /** Region (lower-case). */
    private String userRegion = null;

    /** Current users home directory. */
    private String userHome = null;

    /** Current directory. */
    private String userDir = null;

    /** File name separator (slash or backslash). */
    private String fileSeparator = null;

    /** PATH environment variable separator (colon or semicolon. */
    private String pathSeparator = null;

    /** Host name. */
    private String hostName = null;

    /** Font base URL. */
    private String fontPath = null;

    /** Image base URL. */
    private String imagePath = null;



    /* ********** Constructors ********** */

    /**
     * Constructor.
     * @param a Application name.
     * @param g Application group name.
     * @param u URL to first resource base directory (may be null).
     * @param r Run-Type.
     */
    public DkTool(String a, String g, String u, int r) {
        Properties p = null;
        // Save arguments
        appName = a;
        grpName = g;
        runType = r;
        // Find information about the current user
        try {
            p = System.getProperties();
        } catch(Exception e) {}
        if(p != null) {
            userName = p.getProperty(stt[3]);
            userLanguage = p.getProperty(stt[4]);
            fileSeparator = p.getProperty(stt[12]);
            pathSeparator = p.getProperty(stt[13]);
            if(userLanguage != null) {
                int i = userLanguage.indexOf('_');
                if(i > -1) {
                    userRegion = userLanguage.substring(i + 1);
                    userLanguage = userLanguage.substring(0, i);
                    userRegion = userRegion.toLowerCase();
                }
            }
            if(r == RUN_TYPE_APPLICATION) {
                userHome = p.getProperty(stt[5]);
                userDir = p.getProperty(stt[6]);
            }
        }
        if(fileSeparator == null) { fileSeparator = stt[7]; }
        if(pathSeparator == null) { pathSeparator = stt[11]; }
        // Create preference nodes for application
        if(r == RUN_TYPE_APPLICATION) {
            myprefs = new Preferences[3];
	    myprefs[0] = null;
	    myprefs[1] = null;
	    myprefs[2] = null;
            if(a != null) {
                try {
                    myprefs[0] = Preferences.userRoot().node(a);
                } catch(Exception e) { myprefs[0] = null; }
            }
            if(g != null) {
                try {
                    myprefs[1] = Preferences.userRoot().node(g);
                } catch(Exception e) { myprefs[1] = null; }
            }
            try {
                myprefs[2] = Preferences.userRoot().node(stt[2]);
            } catch(Exception e) { myprefs[2] = null; }
        }
        // Create property set for read-only settings
        props = new Properties();
        mypro = new Properties();
        // Add first directory to resource base list if available
        if(u != null) {
            resourceBaseList = new ArrayList<String>();
            resourceBaseList.add(u);
        }
	if(r == RUN_TYPE_APPLICATION) {
            int hnlevel = 0;
            try {
                NetworkInterface n;
                Enumeration<NetworkInterface> ni =
                NetworkInterface.getNetworkInterfaces();
                boolean cc = true;
                while(ni.hasMoreElements() && cc) {
                    cc = false; n = null;
                    try {
                        n = ni.nextElement();
                    } catch(Exception e) { n = null; }
                    if(n != null) {
                        cc = true;
                        for(InetAddress ia: Collections.list(n.getInetAddresses()))
                        {
                            if(ia.isLoopbackAddress()) {
                                if(hnlevel == 0) {
                                    hnlevel = 1;
                                    hostName = ia.getHostName();
                                }
                            } else {
                                if(ia.isSiteLocalAddress()) {
                                    if(hnlevel <= 1) {
                                        hnlevel = 2;
                                        hostName = ia.getHostName();
                                    }
                                } else {
                                    if(hnlevel <= 2) {
                                        hnlevel = 3;
                                        hostName = ia.getHostName();
                                    }
                                }
                            }
                        }
                    }
                }
            } catch(Exception e) { hostName = null; }
            if(hostName == null) {
                if(r == RUN_TYPE_APPLICATION) {
                    try {
                        hostName = System.getenv(stt[15]);
                    } catch(Exception e) { hostName = null; }
                    if(hostName == null) {
                        try {
                            hostName = System.getenv(stt[16]);
                        } catch(Exception e) { hostName = null; }
                    }
                }
            }
	}
        InputStream cis = openResource(this, stt[17]);
        if(cis != null) {
            propertiesStream(props, cis, (String[])null);
        }
        if(userLanguage == null) {
            userLanguage = props.getProperty(stt[4]);
            if(userLanguage != null) {
                int i = userLanguage.indexOf('_');
                if(i >= 0) {
                    userRegion = userLanguage.substring(i + 1);
                    userLanguage = userLanguage.substring(0, i);
                    userRegion = userRegion.toLowerCase();
                }
            }
        }
        fontPath = props.getProperty(stt[21]);
        imagePath = props.getProperty(stt[22]);
        String resdir = props.getProperty(stt[23]);
        if(resdir != null) {
            if(resourceBaseList == null) {
                resourceBaseList = new ArrayList<String>();
            }
            resourceBaseList.add(resdir);
        }
    }


    /**
     * Constructor.
     * @param a Application name.
     * @param g Application group name.
     * @param u First resource directory URL.
     */
    public DkTool(String a, String g, String u) {
        this(a, g, u, RUN_TYPE_WEBSTART);
    }

    /**
     * Constructor.
     * @param a Application name.
     * @param g Application group name.
     * @param r Run-type (RUN_TYPE_xxx).
     */
    public DkTool(String a, String g, int r) {
        this(a, g, null, r);
    }

    /**
     * Constructor.
     * @param a Application name.
     * @param g Application group name.
     */
    public DkTool(String a, String g) {
        this(a, g, null, RUN_TYPE_WEBSTART);
    }
    
    
    /* ********** Destructor ********** */

    /**
     * Destructor.
     */
    public void cleanup() {
        synchronized(this) {
            if(myprefs != null) {
                try {
                    if(myprefs[0] != null) { myprefs[0].sync(); }
                } catch(Exception e)  {}
                try {
                    if(myprefs[1] != null) { myprefs[1].sync(); }
                } catch(Exception e)  {}
                try {
                    if(myprefs[2] != null) { myprefs[2].sync(); }
                } catch(Exception e)  {}
            }
        }
    }


    /* ********** Information retrieval ********** */


    /**
     * Get current user name.
     * @return  The users logname.
     */
    public String getUserName() {
        return userName;
    }

    /**
     * Get users language.
     * @return  The language.
     */
    public String getUserLanguage() {
        return userLanguage;
    }

    /**
     * Get users language.
     * @return  The language.
     */
    public String getLanguage() {
        return userLanguage;
    }

    /**
     * Get users region.
     * @return  The region converted to lower-case.
     */
    public String getUserRegion() {
        return userRegion;
    }

    /**
     * Get users home directory.
     * @return  The home directory.
     */
    public String getUserHome() {
        return userHome;
    }

    /**
     * Get current directory.
     * @return  The current directory.
     */
    public String getCurrentDirectory() {
        return userDir;
    }

    /**
     * Get host name.
     * @return  The host name.
     */
    public String getHostName() { return hostName; }

    /**
     * Get file separator.
     * @return  The file separator (slash or backslash).
     */
    public String getFileSeparator() { return fileSeparator; }

    /**
     * Get PATH separator.
     * @return  The path separator (colon or semicolon).
     */
    public String getPathSeparator() { return pathSeparator; }

    /**
     * Retrieve run type.
     * @return  The run type RUN_TYPE_xxx.
     */
    public int getRunType() {
        return runType;
    }




    // *********************
    // * String processing *
    // *********************



    /**
     * Get index of string s in string array a.
     * @param a     Array with patterns.
     * @param s     String to search in array.
     * @param cs    Flag: Case-sensitive or not.
     * @return      Index of matched string on success, -1 on error.
     */
    public static int getStringArrayIndex(String[] a, String s, boolean cs) {
        int back = -1;
        int i;
	if((a != null) && (s != null)) {
        for(i = 0; ((i < a.length) && (back == -1)); i++) {
            if(cs) {
                if(s.equals(a[i])) { back = i; }
            } else {
                if(s.equalsIgnoreCase(a[i])) { back = i; }
            }
        }
	}
        return back;
    }



    /**
     * Get index of string s in string array a.
     * @param a Array with patterns.
     * @param s String to search in array.
     * @return  Index of matched string on success, -1 on error.
     */
    public static int getStringArrayIndex(String[] a, String s) {
        int back = getStringArrayIndex(a, s, true);
        return back;
    }



    /**
     * Get index of string s in string array a.
     * @param a     Array to search.
     * @param s     String to search for.
     * @param c     Array of indexes to check.
     * @param cs    Flag: Case sensitive or not.
     * @return      Index in array or -1.
     */
    public static int getStringArrayIndex(
        String[] a, String s, int[] c, boolean cs)
    {
        int back = -1;
        int i;
        int ind;
        if((a != null) && (s != null) && (c != null)) {
            if(c.length > 0) {
                for(i = 0; ((i < c.length) && (back == -1)); i++) {
                    ind = c[i];
                    if((ind >= 0) && (ind < a.length)) {
                        try {
                            if(cs) {
                                if(s.equals(a[ind])) {
                                    back = ind;
                                }
                            } else {
                                if(s.equalsIgnoreCase(a[ind])) {
                                    back = ind;
                                }
                            }
                        } catch(Exception e) {}
                    }
                }
            } else {
                back = getStringArrayIndex(a, s, cs);
            }
        } else {
            back = getStringArrayIndex(a, s, cs);
        }
        return back;
    }



    /**
     * Get index of string s in string array a.
     * @param a     Array to search.
     * @param s     String to search for.
     * @param c     Array of indexes to check.
     * @return      Index in array or -1.
     */
    public static int getStringArrayIndex(String[] a, String s, int[] c) {
        int back = getStringArrayIndex(a, s, c, true);
        return back;

    }



    /**
     * Convert string to boolean.
     * @param s String to convert.
     * @param d Default value.
     * @return  Conversion result or default value.
     */
    public static boolean getBoolean(String s, boolean d) {
        boolean back = d;
	if(s != null) {
        int i = getStringArrayIndex(theBooleanStrings, s);
        if(i > -1) {
            if(i < first_no_boolean) back = true;
            else back = false;
        }
	}
        return back;
    }



    /**
     * Check whether a string is a boolean.
     * @param s String to check.
     * @return  Check result.
     */
    public static boolean isBoolean(String s) {
        boolean back = false;
	if(s != null) {
        if(getStringArrayIndex(theBooleanStrings, s) > -1) {
            back = true;
        }
	}
        return back;
    }



    /**
     * Retrieve integer value from string.
     * @param s String to inspect.
     * @param d Default value.
     * @return  Conversion result or default value.
     */
    public static int getIntValue(String s, int d) {
        int back = d;
        int base = 10;
	if(s != null) {
            if(s.startsWith("0x")) {
                base = 16;
            } else {
                if(s.startsWith("0")) {
                    base = 8;
                }
            }
            try {
                back = Integer.parseInt(s, base);
            } catch(Exception e) { back = d; }
	}
        return back;
    }


    /**
     * Retrieve integer value from string.
     * @param s String to inspect.
     * @return  Conversion result or 0 if conversion failed.
     */
    public static int getIntValue(String s) {
        int back = getIntValue(s, 0);
        return back;
    }



    /**
     * Convert a string into a color.
     * @param s     String to convert.
     * @param dr    Default red.
     * @param dg    Default green.
     * @param db    Default blue.
     * @return      The color retrieved from the string.
     */
    public static Color convertColor(String s, int dr, int dg, int db) {
        int r = dr; int g = dg; int b = db;
        String x;
        if(s != null) {
            try {
                StringTokenizer st = new StringTokenizer(s, ",");
                if(st.hasMoreTokens()) {
                    x = st.nextToken();
                    r = getIntValue(x, dr);
                    if(st.hasMoreTokens()) {
                        x = st.nextToken();
                        g = getIntValue(x, dg);
                        if(st.hasMoreTokens()) {
                            x = st.nextToken();
                            b = getIntValue(x, db);
                        }
                    }
                }
            } catch(Exception e) { r = dr; g = dg; b = db; }
        }
        return(new Color(r, g, b));
    }


    /**
     * Convert a string into a color.
     * @param s     String to convert.
     * @return      The color retrieved from the string.
     */
    public static Color convertColor(String s) {
        Color back = convertColor(s, 255, 255, 255);
        return back;
    }


    /* ********** Preferences ********** */

    /**
     * Set a preferences entry.
     * @param k Entry key.
     * @param v Entry value.
     * @param g Flag: Set for application group too.
     * @param a Flag: Set for all applications too.
     */
    public void setPreferenceString(String k, String v, boolean g, boolean a) {
        String sv = v;
        if(v == null) { sv = stt[8]; }
        if(k != null) {
            synchronized(this) {
                mypro.setProperty(k, sv);
            }
            if(runType == RUN_TYPE_APPLICATION) {
                synchronized(this) {
                    if(myprefs != null) {
                        if(myprefs[0] != null) {
                            try {
                                myprefs[0].put(k, sv);
                            } catch(Exception e) {}
                        }
                        if(g) {
                            if(myprefs[1] != null) {
                                try {
                                    myprefs[1].put(k, sv);
                                } catch(Exception e) {}
                            }
                        }
                        if(a) {
                            if(myprefs[2] != null) {
                                try {
                                    myprefs[2].put(k, sv);
                                } catch(Exception e) {}
                            }
                        }
                    }
                }
            }
        }
    }
    

    /**
     * Set preferences entry.
     * @param k Entry key.
     * @param v Entry name.
     */
    public void setPreferenceString(String k, String v) {
        setPreferenceString(k, v, false, false);
    }



    /**
     * Set multiple preference entries.
     * @param k Entry keys.
     * @param v Entry values.
     * @param g Flag: Set for application group too.
     * @param a Flag: Set as general defaults too.
     */
    public void setMultiplePreferenceStrings(
            String[] k, String[] v,
            boolean g, boolean a
            )
    {
        int max;
        if((k != null) && (v != null)) {
            max = k.length;
            if(max > v.length) max = v.length;
            for(int i = 0; i < max; i++) {
                setPreferenceString(k[i], v[i], g, a);
            }
        }
    }


    /**
     * Set multiple preference entries.
     * @param k Entry keys.
     * @param v Entry values.
     */
    public void setMultiplePreferenceStrings(String[] k, String[] v) {
        setMultiplePreferenceStrings(k, v, false, false);
    }


    /**
     * Get preference/property entry.
     * @param k Entry key.
     * @param d Entry value.
     * @param u Flag: Use user entries (preferences) too.
     * @return  The setting
     */
    private String internalPreferenceString(String k, String d, boolean u) {
        String back = null;
	if(u) {
            synchronized(this) {
                back = mypro.getProperty(k);
            }
            if(runType == RUN_TYPE_APPLICATION) {
                for(int i = 0; i < 3; i++) {
                    if(back == null) {
                        synchronized(this) {
                            if(myprefs != null) {
                                if(myprefs[i] != null) {
                                    try {
                                        back = myprefs[i].get(k, d);
                                    } catch(Exception e) { back = null; }
                                }
                            }
                        }
                    }
                }
            }
	}
        if(back == null) {
            synchronized(this) {
                back = props.getProperty(k);
            }
        }
        if(back == null) {
            try {
                back = System.getProperties().getProperty(k);
            } catch(Exception e) { back= null; }
        }
        if(back == null) {
            back = d;
        }
        return back;
    }


    /**
     * Get preferences entry.
     * @param k Entry key.
     * @param d Default value.
     * @return  Entry value or default value.
     */
    public String getPreferenceString(String k, String d) {
        String back = internalPreferenceString(k, d, true);
	return back;
    }


    /**
     * Get property entry.
     * @param k Entry key.
     * @param d Default value.
     * @return  Entry value or default value.
     */
    public String getPropertyString(String k, String d) {
        String back = internalPreferenceString(k, d, false);
	return back;
    }


    /**
     * Get multiple preference entries.
     * @param k Key array.
     * @param v Value array.
     */
    public void getMultiplePreferenceStrings(String[] k, String[] v) {
        int max;
        if((k != null) && (v != null)) {
            max = k.length;
            if(v.length < max) max = v.length;
            for(int i = 0; i < max; i++) {
                if(k[i] != null) {
                    v[i] = getPreferenceString(k[i], v[i]);
                }
            }
        }
    }



    /**
     * Set preferences entry.
     * @param k Entry key.
     * @param v Entry value.
     * @param g Flag: Set for application group too.
     * @param a Flag: Set as general defaults too.
     */
    public void setPreferenceInt(String k, int v, boolean g, boolean a) {
        if(k != null) {
            setPreferenceString(k, (new Integer(v)).toString(), g, a);
        }
    }


    /**
     * Set preferences entry.
     * @param k Entry key.
     * @param v Entry value.
     */
    public void setPreferenceInt(String k, int v) {
        setPreferenceInt(k, v, false, false);
    }



    /**
     * Get preferences entry.
     * @param k Entry key.
     * @param d Default value.
     * @return  Entry value or default value.
     */
    public int getPreferenceInt(String k, int d) {
        int back = d;
        try {
            String v = getPreferenceString(k, null);
            back = Integer.parseInt(v);
        } catch(Exception e) { back = d; }
        return back;
    }


    /**
     * Set multiple preferences entries.
     * @param k Key array.
     * @param v Value array.
     * @param g Flag: Set entries for application group too.
     * @param a Flag: Set as default values.
     */
    public void setMultiplePreferenceInts(
            String[] k, int[] v, boolean g, boolean a
            )
    {
        int max;
        if((k != null) && (v != null)) {
            max = k.length;
            if(max > v.length) max = v.length;
            for(int i = 0; i < max; i++) {
                setPreferenceInt(k[i], v[i], g, a);
            }
        }
    }

    /**
     * Set multiple preferences entries.
     * @param k Keys array.
     * @param v Values array.
     */
    public void setMultiplePreferenceInts(String[] k, int[] v) {
        setMultiplePreferenceInts(k, v, false, false);
    }


    /**
     * Get multiple preferences entries.
     * @param k Keys array.
     * @param v Values array.
     */
    public void getMultiplePreferenceInts(String[] k, int[] v) {
        int max;
        if((k != null) && (v != null)) {
            max = k.length;
            if(max > v.length) max = v.length;
            for(int i = 0; i < max; i++) {
                v[i] = getPreferenceInt(k[i], v[i]);
            }
        }
    }


    /**
     * Set preference entry.
     * @param k Entry key.
     * @param v Entry value.
     * @param g Flag: Set for application group too.
     * @param a Flag: Set as default.
     */
    public void setPreferenceBool(String k, boolean v, boolean g, boolean a) {
        if(k != null) {
            String sv = theBooleanStrings[v ? 3 : 10];
            setPreferenceString(k, sv, g, a);
        }
    }


    /**
     * Get preference entry.
     * @param k Entry key.
     * @param d Default value.
     * @return  Entry value or default value.
     */
    public boolean getPreferenceBool(String k, boolean d) {
        boolean back = d;
        try {
            String v = getPreferenceString(k, null);
            int i = getStringArrayIndex(theBooleanStrings, v);
            if(i > -1) {
                if(i < first_no_boolean) {
                    back = true;
                } else {
                    back = false;
                }
            }
        } catch(Exception e) { back = d; }
        return back;
    }


    /**
     * Set multiple preferences entries.
     * @param k Keys array.
     * @param v Values array.
     * @param g Flag: Set entries for application group too.
     * @param a Flag: Set as default for all applications.
     */
    public void setMultiplePreferenceBools(
            String[] k, boolean[] v, boolean g, boolean a
            )
    {
        int max;
        if((k != null) && (v != null)) {
            max = k.length;
            if(max > v.length) max = v.length;
            for(int i = 0; i < max; i++) {
                setPreferenceString(k[i],theBooleanStrings[v[i] ? 3 : 10],g,a);
            }
        }
    }



    /**
     * Get multiple preferences entries.
     * @param k Keys array.
     * @param v Values array.
     */
    public void getMultiplePreferenceBools(String[] k, boolean[] v) {
        int max;
        if((k != null) && (v != null)) {
            max = k.length;
            if(max > v.length) max = v.length;
            for(int i = 0; i < max; i++) {
                v[i] = getPreferenceBool(k[i], v[i]);
            }
        }
    }



    /**
     * Set multiple preferences entries.
     * @param k Keys array.
     * @param v Values array.
     */
    public void setMultiplePreferenceBools(String[] k, boolean[] v) {
        setMultiplePreferenceBools(k, v, false, false);
    }


    /**
     * Flush preferences to permanent storage.
     */
    public void flushPreferences() {
        if(runType == RUN_TYPE_APPLICATION) {
	    synchronized(this) {
	        if(myprefs != null) {
                    try {
                        if(myprefs[0] != null) { myprefs[0].sync(); }
                    } catch(Exception e) {}
                    try {
                        if(myprefs[1] != null) { myprefs[1].sync(); }
                    } catch(Exception e) {}
                    try {
                        if(myprefs[2] != null) { myprefs[2].sync(); }
                    } catch(Exception e) {}
		}
	    }
	}
    }


    /* ********** Resource search ********** */

    /**
     * Get input stream for URL.
     * @param fn    URL.
     * @return  Input stream or null.
     */
    private InputStream streamForUrl(String fn) {
        InputStream back = null;
        try {
            URL url = new URL(fn);
            back = url.openStream();
        } catch(Exception e) { back = null; }
        return back;
    }


    /**
     * Get input stream for file name.
     * @param fn    File name.
     * @return  Input stream or null.
     */
    private InputStream streamForFilename(String fn) {
        InputStream back = null;
        try {
            File f = new File(fn);
            if(f.exists()) {
                if(f.isFile()) {
                    back = new FileInputStream(f);
                }
            }
        } catch(Exception e) { back = null; }
        return back;
    }


    /**
     * Get input stream for resource in *.jar file.
     * @param o         Object originating the request.
     * @param fn        File name.
     * @param sslash    Flag: File name start with slash.
     * @return          Input stream or null.
     */
    private InputStream streamForResource(Object o, String fn, boolean sslash)
    {
        InputStream back = null;
	URL u;
        try {
            back = o.getClass().getClassLoader().getResourceAsStream(fn);
        } catch(Exception e) { back = null; }
        if(back == null) {
            if(!sslash) {
                try {
                    back = o.getClass().getClassLoader().getResourceAsStream(
                        stt[7] + fn
                    );
                } catch(Exception e) {
                    back = null;
                }
            }
        }
	if(back == null) {
	    try {
	        u = o.getClass().getResource(fn);
		back = u.openStream();
	    } catch(Exception e) { back = null; }
	}
	if(back == null) {
	    if(!sslash) {
	        try {
		    u = o.getClass().getResource(stt[7] + fn);
		    back = u.openStream();
		} catch(Exception e) { back = null; }
	    }
	}
        return back;
    }


    /**
     * Get input stream for a named object.
     * @param o     Object originating the request.
     * @param fn    File name.
     * @return      Input stream or null.
     */
    private InputStream streamForName(Object o, String fn) {
        InputStream back = null;
        if(fn.startsWith(stt[9])) {         // http:
            back = streamForUrl(fn);
        } else {
            if(fn.startsWith(stt[10])) {    // file:
                if(runType == RUN_TYPE_APPLICATION) {
                    back = streamForUrl(fn);
                }
            } else {
                if(fn.startsWith(stt[7])) { // /
                    if(runType == RUN_TYPE_APPLICATION) {
                        back = streamForFilename(fn);
                    }
                } else {                    // in jar
		    back = streamForResource(o, fn, false);
                }
            }
        }
        return back;
    }


    /**
     * Add a directory (file name or URL) to the list of resource bases.
     * @param s Directory to add.
     */
    public void addResourceBase(String s) {
        if(s != null) {
	    synchronized(this) {
	        if(resourceBaseList == null) {
		    resourceBaseList = new ArrayList<String>();
		}
		resourceBaseList.add(s);
	    }
	}
    }


    /**
     * Get input stream to read from a named resource.
     * @param o     Object originating the request.
     * @param ed    Extra directory (may be null).
     * @param fn    File name (must not contain http:// or file://).
     * @param loc   Flag: Search for localized versions.
     * @return      Input stream or null.
     */
    private InputStream searchForResource(Object o, String ed, String fn, boolean loc) {
        InputStream back = null;
        Iterator<String> rbli = null;
        boolean cc;
        String dirname;
        String filename;
        boolean startsWithSlash = fn.startsWith(stt[7]);
	/* Check *.jar file containing the application first */
        if(back == null) {
            if(loc) {
                if(userLanguage != null) {
                    if(userRegion != null) {
                        filename = userLanguage + stt[7] + userRegion
                            + stt[startsWithSlash ? 8 : 7]+ fn;
                        back = streamForResource(o, filename, false);
                    }
                    if(back == null) {
                        filename = userLanguage + stt[startsWithSlash ? 8 : 7]+ fn;
                        back = streamForResource(o, filename, false);
                    }
                }
            }
            if(back == null) {
                back = streamForResource(o, fn, startsWithSlash);
            }
        }
	/* Extra directory (if any) */
	if(back == null) {
	    if(ed != null) {
                dirname = ed;
                if(loc) {
                    if(userLanguage != null) {
                        if(userRegion != null) {
                            filename = dirname + stt[7] + userLanguage + stt[7]
                                + userRegion
                                + stt[startsWithSlash ? 8 : 7] + fn;
                            back = streamForName(o, filename);
                        }
                        if(back == null) {
                            filename = dirname + stt[7] + userLanguage
                                   + stt[startsWithSlash ? 8 : 7] + fn;
                            back = streamForName(o, filename);
                        }
                    }
                }
                if(back == null) {
                    filename = dirname + stt[startsWithSlash ? 8 : 7] + fn;
                    back = streamForName(o, filename);
                }
	    }
	}
	/* Walk through resource base directories */
	if(back == null) {
            synchronized(this) {
                if(resourceBaseList != null) {
                    try {
                        rbli = resourceBaseList.iterator();
                    } catch(Exception e) { rbli = null; }
                }
            }
            cc = true;
            while(cc && (back == null)) {
                dirname = null; cc = false;
                try {
                    dirname = rbli.next();
                } catch(Exception e) { dirname = null; }
                if(dirname != null) {
                    cc = true;
                    if(loc) {
                        if(userLanguage != null) {
                            if(userRegion != null) {
                                filename = dirname + stt[7] + userLanguage + stt[7]
                                           + userRegion
                                           + stt[startsWithSlash ? 8 : 7] + fn;
                                back = streamForName(o, filename);
                            }
                            if(back == null) {
                                filename = dirname + stt[7] + userLanguage
                                           + stt[startsWithSlash ? 8 : 7] + fn;
                                back = streamForName(o, filename);
                            }
                        }
                    }
                    if(back == null) {
                        filename = dirname + stt[startsWithSlash ? 8 : 7] + fn;
                        back = streamForName(o, filename);
                    }
                }
            }
	}
	/* Finally check current directory */
        if(back == null) {
            if(runType == RUN_TYPE_APPLICATION) {
		dirname = userDir;
		if(dirname != null) {
		    if(loc) {
		        if(userLanguage != null) {
			    if(userRegion != null) {
                            filename = dirname + stt[7] + userLanguage + stt[7]
                                       + userRegion
                                       + stt[startsWithSlash ? 8 : 7] + fn;
                            back = streamForFilename(filename);
			    }
			    if(back == null) {
                            filename = dirname + stt[7] + userLanguage
                                       + stt[startsWithSlash ? 8 : 7] + fn;
                            back = streamForFilename(filename);
			    }
			}
		    }
		    if(back == null) {
		        filename = dirname + stt[startsWithSlash ? 8 : 7] + fn;
			back = streamForFilename(filename);
		    }
		}
            }
        }
        return back;
    }


    /**
     * Get input stream to read from a named resource.
     * @param o     Object originating the request.
     * @param fn    File name.
     * @param loc   Flag: Search for localized versions.
     * @return      Input stream or null.
     */
    private InputStream searchForResource(Object o, String fn, boolean loc) {
        InputStream back = searchForResource(o, null, fn, loc);
	return back;
    }



    /**
     * Get input stream for a named resource, prefer localized version.
     * @param o Object originating the request.
     * @param n File name.
     * @return  Input stream or null.
     */
    public InputStream openLocalizedResource(Object o, String n) {
        InputStream back = null;
        if((o != null) && (n != null)) {
            back = searchForResource(o, n, true);
        }
        return back;
    }


    /**
     * Get input stream for a named resource.
     * @param o Object originating the request.
     * @param n File name.
     * @return  Input stream or null.
     */
    public InputStream openResource(Object o, String n) {
        InputStream back = null;
	if((o != null) && (n != null)) {
	    back = searchForResource(o, n, false);
	}
	return back;
    }



    /**
     * Get buffered image.
     * @param	o	Object originating the request.
     * @param	fn	File name.
     * @return	Image or null.
     */
    public BufferedImage findBufferedImage(Object o, String fn) {
        BufferedImage back = null;
	InputStream is = null;
	if((o != null) && (fn != null)) {
	    if(fn.startsWith(stt[9])) {
	        try {
	            is = (new URL(fn)).openStream();
	        } catch(Exception e) { is = null; }
	    } else {
	        if(fn.startsWith(stt[10])) {
	            if(runType == RUN_TYPE_APPLICATION) {
		        try {
		            is = (new URL(fn)).openStream();
		        } catch(Exception e) { is = null; }
		    }
	        } else {
	            String idir = null;
		    synchronized(this) { idir = imagePath; }
	            is = searchForResource(o, idir, fn, true);
	        }
	    }
	    if(is != null) {
	        try {
	            back = ImageIO.read(is);
	        } catch(Exception e) { back = null; }
	        try { is.close(); } catch(Exception e) {}
	    }
	}
	return back;
    }


    /**
     * Create image icon.
     * @param	o	Object.
     * @param	fn	File name.
     * @return	Image icon or null.
     */
    public ImageIcon findImageIcon(Object o, String fn) {
        ImageIcon back = null;
	if((o != null) && (fn != null)) {
	    BufferedImage bi = findBufferedImage(o, fn);
	    if(bi != null) {
	        try {
		    back = new ImageIcon(bi);
		} catch(Exception e) { back = null; }
	    }
	    if(back == null) {
	        try {
		    back = new ImageIcon(o.getClass().getResource(fn));
		} catch(Exception e) { back = null; }
	    }
	}
	return back;
    }



    /**
     * Fint font.
     * @param	o	Object originating the request.
     * @param	ffn	Font file names.
     * @param	fn	Font names.
     * @param	fb	Fallback font.
     * @param	fe	Features of the font.
     * @param	fs	Font size.
     * @return	Found font or null.
     */
    public
    Font
    findFont(
        Object o, String[] ffn, String[] fn, String fb, int fe, float fs
    )
    {
        Font back = null;
	Font xf = null;
	InputStream is = null;
	int i;
	int ft;
	if((o != null) && (ffn != null)) {
	    for(i = 0; ((i < ffn.length) && (back == null)); i++) {
	        is = null; xf = null;
		ft = Font.TYPE1_FONT;
                if(ffn[i] != null) {
		    if(ffn[i].endsWith(stt[24]) || ffn[i].endsWith(stt[25])) {
		        ft = Font.TRUETYPE_FONT;
		    }
		    if(ffn[i].startsWith(stt[9])) {
		        try {
		            is = (new URL(ffn[i])).openStream();
			    xf = Font.createFont(ft, is);
		        }
		        catch(Exception e) {
		            xf = null;
		        }
		        finally {
		            try { is.close(); } catch(Exception e) {}
		        }
		    } else {
		        if(ffn[i].startsWith(stt[10])) {
		            if(runType == RUN_TYPE_APPLICATION) {
			        try {
			            is = (new URL(ffn[i])).openStream();
				    xf = Font.createFont(ft, is);
			        }
			        catch(Exception e) {
			            xf = null;
			        }
			        finally {
			            try { is.close(); } catch(Exception e) {}
			        }
			    }
		        } else {
		            String fp;
			    synchronized(this) { fp = fontPath; }
		            try {
			        is = searchForResource(o, fp, ffn[i], false);
			        xf = Font.createFont(ft, is);
			    }
			    catch(Exception e) {
			        xf = null;
			    }
			    finally {
			        try { is.close(); } catch(Exception e) {}
			    }
		        }
		    }
		    if(xf != null) {
		        try {
		            back = xf.deriveFont(fs);
		        } catch(Exception e) { back = null; }
		    }
		}
	    }
	}
	if(fn != null) {
	    for(i = 0; ((i < fn.length) && (back == null)); i++) {
	        if(fn[i] != null) {
		    try {
		        xf = new Font(fn[i], fe, (int)fs);
			back = xf.deriveFont(fs);
		    } catch(Exception e) { back = null; }
		}
	    }
	}
	if(back == null) {
	    if(fb != null) {
	        try {
		    xf = new Font(fb, fe, (int)fs);
		    back = xf.deriveFont(fs);
		} catch(Exception e) { back = null; }
	    }
	}
	return back;
    }



    /**
     * Fint font.
     * @param	o	Object originating the request.
     * @param	ffn	Font file names.
     * @param	fn	Font names.
     * @param	fb	Fallback font.
     * @param	fe	Features of the font.
     * @param	fs	Font size.
     * @return	Found font or null.
     */
    public
    Font
    findFont(
        Object o, String[] ffn, String[] fn, String fb, int fe, double fs
    )
    {
        Font back = findFont(o, ffn, fn, fb, fe, (float)fs);
	return back;
    }



    /**
     * Get a string array from an input stream.
     * @param is    Input stream to process.
     * @param csn   Character set name ("UTF-8" or "ISO-LATIN-8859-1").
     * @return      String array or null.
     */
    private String[] getArrayFromInputStream(InputStream is, String csn) {
        String[] back = null;
        InputStreamReader isr = null;
        BufferedReader br = null;
        ArrayList<String> al = new ArrayList<String>();
        try {
            if(csn != null) { isr = new InputStreamReader(is, csn); }
            else { isr = new InputStreamReader(is); }
            br = new BufferedReader(isr);
            String lineBuffer;
            boolean cc;
            cc = true;
            while(cc) {
                lineBuffer = null; cc = false;
                try {
                    lineBuffer = br.readLine();
                } catch(Exception e) { lineBuffer = null; }
                if(lineBuffer != null) {
                    cc = true;
                    if(!lineBuffer.startsWith("#")) {
                        lineBuffer = processBackslashEscapedString(lineBuffer);
                        al.add(lineBuffer);
                    }
                }
            }
            Object[] oa = al.toArray();
            back = new String[oa.length];
            for(int i = 0; i < oa.length; i++) {
                back[i] = new String((String)(oa[i]));
            }
        }
        catch(Exception e) {
            back = null;
        }
        finally {
            try { br.close(); } catch(Exception e) {}
            try { isr.close(); } catch(Exception e) {}
        }
        return back;
    }


    /**
     * Get a localized array of strings from a named file.
     * @param o     Object originating the request.
     * @param fn    File name.
     * @param csn   Character set name.
     * @return      String array or null.
     */
    public String[] getLocalizedArray(Object o, String fn, String csn) {
        String[] back = null;
        if((o != null) && (fn != null)) {
            InputStream is = openLocalizedResource(o, fn);
            if(is != null) {
                back = getArrayFromInputStream(is, csn);
                try { is.close(); } catch(Exception e) {}
            }
        }
        return back;
    }

    /**
     * Get a localized array of strings from a named file.
     * @param o     Object originating the request.
     * @param fn    File name.
     * @return      String array or null.
     */
    public String[] getLocalizedArray(Object o, String fn) {
        String[] back = getLocalizedArray(o, fn, null);
        return back;
    }


    /**
     * Get font path.
     * @return  Font base directory.
     */
    public String getFontPath() {
        String back;
        synchronized(this) { back = fontPath; }
        return back;
    }


    /**
     * Get images path.
     * @return  Images base directory.
     */
    public String getImagePath() {
        String back;
        synchronized(this) { back = imagePath; }
        return back;
    }



    /**
     * Construct one large string from an input stream contents.
     * @param is    Input stream to process.
     * @param csn   Character set name ("UTF-8" or "ISO-8859-1" or null).
     * @return      A string containing the file contents.
     */
    public String getStringFromInputStream(InputStream is, String csn) {
        String back = null;
        StringBuilder sb = new StringBuilder();
        InputStreamReader isr = null;
        BufferedReader br = null;
        try {
            if(csn != null) { isr = new InputStreamReader(is, csn); }
            else { isr = new InputStreamReader(is); }
            br = new BufferedReader(isr);
            String lineBuffer;
            boolean cc = true;
            while(cc) {
                lineBuffer = null; cc = false;
                try {
                    lineBuffer = br.readLine();
                } catch(Exception e) { lineBuffer = null; }
                if(lineBuffer != null) {
                    cc = true;
                    sb.append(lineBuffer);
		    sb.append('\n');
                }
            }
        }
        catch(Exception e) {
        }
        finally {
            try { br.close(); } catch(Exception e) {}
            try { isr.close(); } catch(Exception e) {}
        }
        back = new String(sb);
        return back;
    }


    /**
     * Construct one large string from an input stream contents.
     * @param is    Input stream to process.
     * @return      A string containing the file contents.
     */
    public String getStringFromInputStream(InputStream is) {
        String back = getStringFromInputStream(is, null);
        return back;
    }


    /**
     * Get file contents as one string.
     * @param o     Object.
     * @param fn    File name.
     * @param csn   Character set name.
     * @return      File contents as one large string.
     */
    public String getLocalizedString(Object o, String fn, String csn) {
        String back = null;
        InputStream is = null;
        is = openLocalizedResource(o, fn);
        if(is != null) {
            back = getStringFromInputStream(is, csn);
            try { is.close(); } catch(Exception e) {}
        }
        return back;
    }


    /**
     * Get file contents as one string.
     * @param o     Object.
     * @param fn    File name.
     * @return      File contents as one large string.
     */
    public String getLocalizedString(Object o, String fn) {
        String back = getLocalizedString(o, fn, null);
        return back;
    }


    /**
     * Process a backslash-escaped string.
     * @param s Source string.
     * @return  Processed string.
     */
    public String processBackslashEscapedString(String s) {
        String back = null;
        char c;
        StringBuffer sb = new StringBuffer();
        boolean lastWasBackslash = false;
        if(s != null) {
            for(int i = 0; i < s.length(); i++) {
                c = s.charAt(i);
                if(lastWasBackslash) {
                    switch(c) {
                        case 'n': {
                            sb.append('\n');
                        } break;
                        case 'r': {
                            sb.append('\r');
                        } break;
                        case 't': {
                            sb.append('\t');
                        } break;
                        case '\r': case '\n': {
                            i = s.length();
                        } break;
                        default: {
                            sb.append(c);
                        } break;
                    }
                } else {
                    if(c == '\\') {
                        lastWasBackslash = true;
                    } else {
                        lastWasBackslash = false;
                        if((c == '\r') || (c == '\n')) {
                            i = s.length();
                        } else {
                            sb.append(c);
                        }
                    }
                }
            }
            back = new String(sb);
        }
        return back;
    }




    /**
     * Trim and squeeze a string (return text only).
     * Leading and trailing whitespaces are removed, other sequences
     * of whitespaces are replaced by one space.
     * @param s Original string.
     * @return  Trimmed and squeezed string or null.
     */
    public String trimString(String s) {
        String back = null;
        boolean contentsFound = false;
        boolean spaceDelayed = false;
        int i;
        char c;
        StringBuilder sb = new StringBuilder();
	if(s != null) {
            for(i = 0; i < s.length(); i++) {
                c = s.charAt(i);
                switch(c) {
                    case ' ': case '\t': case '\r': case '\n': {
                        spaceDelayed = true;
                    } break;
                    default: {
                        if(spaceDelayed) {
                            sb.append(' ');
                        }
                        sb.append(c);
                        spaceDelayed = false;
                        contentsFound = true;
                    } break;
                }
            }
            if(contentsFound) {
                back = new String(sb);
            }
	}
        return back;
    }


    /**
     * Check whether s starts with one of the strings in pre.
     * @param pre   Array of prefixes to check.
     * @param s     String to check.
     * @return      Check result.
     */
    private boolean startsWithOneFromArray(String[] pre, String s) {
        boolean back = false;
        if((pre != null) && (s != null)) {
            for(int i = 0; ((i < pre.length) && (!back)); i++) {
                if(pre[i] != null) {
                    if(s.startsWith(pre[i])) {
                        back = true;
                    }
                }
            }
        } else {
	    if(pre == null) { back = true; }
	}
        return back;
    }


    /**
     * Apply a properties stream to a Properties object.
     * @param p     Properties object.
     * @param isr   Input stream to process.
     * @param pre   Array of property name beginnings.
     */
    @SuppressWarnings("unchecked")
    private void propertiesStream(Properties p, InputStream isr, String[] pre)
    {
        InputStreamReader iread = null;
        BufferedReader br = null;
        String su;                      // Scope user
        String sh;                      // Scope host
        String sa;                      // Scope application
        boolean lastWasScope = false;   // Flag: Previous line was a scope line
        int i;
        String inputLine;               // One line to process
        String ks;                      // Key string
        String kv;                      // Value string
        int tmi;                        // Index of current map
        TreeMap<String,String> tm;
        Object[] tma = new Object[8];   // Maps for scopes
    	if((p != null) && (isr != null)) {
            for(i = 0; i < 8; i++) {
                tma[i] = new TreeMap<String,String>();
            }
            tmi = 0; tm = (TreeMap<String,String>)(tma[0]);
            try {
                iread = new InputStreamReader(isr);
                br = new BufferedReader(iread);
                boolean cc = true; lastWasScope = false;
                while(cc) {
                    cc = false; inputLine = null;
                    try { inputLine = br.readLine(); }
                    catch(Exception e) { inputLine = null;}
                    if(inputLine != null) {
                        cc = true;
                        if(!inputLine.startsWith(stt[20])) {
                            inputLine = trimString(inputLine);
                            if(inputLine != null) {
                                if(inputLine.startsWith(stt[18])) {
                                    if(!lastWasScope) {
                                        tmi = -1; tm = null;
                                    } lastWasScope = true;
                                    inputLine = inputLine.substring(1);
                                    i = inputLine.indexOf(']');
                                    if(i > 0) {
                                        su = null; sh = null; sa = null;
                                        inputLine = inputLine.substring(0, i);
                                        i = inputLine.indexOf('/');
                                        if(i >= 0) {
                                            su = inputLine.substring(0,i);
                                            inputLine= inputLine.substring(i+1);
                                            i = inputLine.indexOf('/');
                                            if(i >= 0) {
                                                sh = inputLine.substring(0,i);
                                                sa = inputLine.substring(i+1);
                                            } else {
                                                sh = inputLine;
                                            }
                                        } else {
                                            su = inputLine;
                                        }
                                        i = -1;
                                        if(userName != null) {
                                            if(su != null) {
                                                if(su.equals(userName)) {
                                                    i = 4;
                                                } else {
                                                    if(su.equals(stt[19])) {
                                                        i = 0;
                                                    }
                                                }
                                            }
                                        } else {
                                            if(su != null) {
                                                if(su.equals(stt[19])) {
                                                    i = 0;
                                                }
                                            }
                                        }
                                        if(i > -1) {
                                            if(sh != null) {
                                                if(!sh.equals(stt[19])) {
                                                    if(hostName != null) {
                                                        if(sh.equals(hostName))
                                                        {
                                                            i += 2;
                                                        } else {
                                                            i = -1;
                                                        }
                                                    } else {
                                                        i = -1;
                                                    }
                                                }
                                            }
                                        }
                                        if(i > -1) {
                                            if(sa != null) {
                                                if(!sa.equals(stt[19])) {
                                                    if(appName != null) {
                                                        if(sa.equals(appName)) {
                                                            i += 1;
                                                        } else {
                                                            i = -1;
                                                        }
                                                    } else {
                                                        i = -1;
                                                    }
                                                }
                                            }
                                        }
                                        if(i > tmi) {
                                            tmi = i;
                                            if((tmi >= 0) && (tmi < 8)) {
                                                tm =
                                                (TreeMap<String,String>)(tma[tmi]);
                                            }
                                        }
                                    }
                                } else {
                                    lastWasScope = false;
                                    if((tmi > -1) && (tmi < 8)) {
                                        i = inputLine.indexOf('=');
                                        if(i >= 0) {
                                            ks = inputLine.substring(0,i);
                                            kv = inputLine.substring(i+1);
                                            if(tm != null) {
                                                try {
                                                    tm.put(ks, kv);
                                                } catch(Exception e) {}
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
            catch(Exception e) { e.printStackTrace(); }
            finally {
                try { br.close(); } catch(Exception e) {}
                try { iread.close(); } catch(Exception e) {}
            }
            // All input handled, now add properties.
            for(i = 0; i < 8; i++) {
                tm = (TreeMap<String,String>)(tma[i]);
                for(String myk: tm.keySet()) {
                    try {
                        String myv = (String)tm.get(myk);
			if(myv != null) {
                            p.put(myk, myv);
			}
                    } catch(Exception e) {}
                }
            }
	}
    }


    /**
     * Apply a properties stream to a Properties object.
     * @param p     Properties object.
     * @param isr   Input stream to process.
     * @param pre   Beginning of property names of interest.
     */
    private void propertiesStream(Properties p, InputStream isr, String pre) {
        if(pre != null) {
	    String[] newpre = new String[1];
	    newpre[0] = pre;
	    propertiesStream(p, isr, newpre);
	} else {
            propertiesStream(p, isr, (String[])null);
        }
    }

    /**
     * Convert a text string to PS encoding, so we can insert it into
     * a PS file.
     * @param s String to convert.
     * @return  Conversion result.
     */
    public static String encodeForPostscript(String s) {
      String back;
      StringBuilder res = new StringBuilder();
      int i;
      char c;
      int ic;
      String x;
      if(s != null) {
      for(i = 0; i < s.length(); i++) {
          c = s.charAt(i);
          if(psAllowedCharacters.indexOf((int)c) >= 0) {
              res.append(c);
          } else {
              switch(c) {
                  case '\n': { res.append("\\n"); } break;
                  case '\t': { res.append("\\t"); } break;
                  case '\f': { res.append("\\f"); } break;
                  case '\r': { res.append("\\r"); } break;
                  case '\b': { res.append("\\b"); } break;
                  case '\\': { res.append("\\\\"); } break;
                  default: {
                      ic = (int)c;
                      if((ic >= 0) && (ic <= 255)) {
                          x = Integer.toOctalString(ic);
                          switch(x.length()) {
                              case 0: { res.append("\\000"); } break;
                              case 1: {
                                  res.append("\\00"); res.append(x);
                              } break;
                              case 2: {
                                  res.append("\\0"); res.append(x);
                              } break;
                              case 3: {
                                  res.append("\\"); res.append(x);
                              } break;
                          }
                      }
                  } break;
              }
          }
      }
      }
      back = new String(res);
      return back;
    }



    /**
     * Find position of file name suffix.
     * @param fn    File name
     * @return Index of the suffix dot or -1.
     */
    public int getFilenameSuffixPosition(String fn) {
        int back = -1;
        char c;
	if(fn != null) {
        int l = fn.length();
        boolean sepbs = fileSeparator.equals(stt[14]);
        for(int i = 0; i < l; i++) {
            c = fn.charAt(i);
            if(c == '.') {
                back = i;
            } else {
                if(sepbs) {
                    if(c == '\\') { back = -1; }
                } else {
                    if(c == '/') { back = -1; }
                }
            }
        }
	}
        return back;
    }


    /**
     * Find file name suffix (including the dot).
     * @param fn    File name to inspect.
     * @return  File name suffix or null.
     */
    public String getFilenameSuffix(String fn) {
        String back = null;
	if(fn != null) {
        int posdot = getFilenameSuffixPosition(fn);
        if(posdot >= 0) {
            back = fn.substring(posdot);
        }
	}
        return back;
    }


    /**
     * Replace file suffixes by a new suffix (or add new suffix).
     * @param fn    Original file name.
     * @param ns    New suffix to append.
     * @return      The modified file name
     */
    public String replaceSuffix(String fn, String ns) {
        String back = null;
	if((fn != null) && (ns != null)) {
        int sp = getFilenameSuffixPosition(fn);
        if(sp > -1) {
            String bn = fn.substring(0, sp);
            back = new String(bn + ns);
        } else {
            back = new String(fn + ns);
        }
	}
        return back;
    }

    /**
     * Return string for HTML encoding of one character.
     * @param c Character to encode.
     * @return  HTML representation of c.
     */
    private String htmlEncodedCharacter(char c) {
        StringBuilder sb = new StringBuilder();
        sb.append("&#x");
        if(c > 0xFFFFFF) {
            sb.append(hexCharacters[(c >> 28) & 0x0F]);
            sb.append(hexCharacters[(c >> 24) & 0x0F]);
        }
        if(c > 0xFFFF) {
            sb.append(hexCharacters[(c >> 20) & 0x0F]);
            sb.append(hexCharacters[(c >> 16) & 0x0F]);
        }
        if(c > 0xFF) {
            sb.append(hexCharacters[(c >> 12) & 0x0F]);
            sb.append(hexCharacters[(c >>  8) & 0x0F]);
        }
        sb.append(hexCharacters[(c >> 4) & 0x0F]);
        sb.append(hexCharacters[c        & 0x0F]);
        sb.append(";");
        String back = new String(sb);
        return back;
    }



    /**
     * Create HTML representation of a string.
     * @param s Original string.
     * @return  HTML code to show the string.
     */
    public String encodeToHtml(String s) {
        String back = null;
        char c;
        StringBuilder sb = new StringBuilder();
	if(s != null) {
        int l = s.length();
        for(int i = 0; i < l; i++) {
            c = s.charAt(i);
            if((c >= 'a') && (c <= 'z')) {
                sb.append(c);
            } else {
                if((c >= 'A') && (c <= 'Z')) {
                    sb.append(c);
                } else {
                    if(c == '0') {
                        sb.append(c);
                    } else {
                        if((c >= '1') && (c <= '9')) {
                            sb.append(c);
                        } else {
                            switch(c) {
                                case '+':
                                case '-':
                                case '!':
                                case '/':
                                case '(':
                                case ')':
                                case '[':
                                case ']':
                                case '{':
                                case '}':
                                case '=':
                                case '?':
                                case '*':
                                case '~':
                                case '#':
                                case '_':
                                case '.':
                                case ':':
                                case ',':
                                case ';':
                                case '|':
                                {
                                    sb.append(c);
                                } break;
                                case '<': {
                                    sb.append("&lt;");
                                } break;
                                case '>': {
                                    sb.append("&gt;");
                                } break;
                                case ' ': {
                                    sb.append("&nbsp;");
                                } break;
                                default: {
                                    // sb.append(String.format("&#x%04x;", c));
                                    sb.append(htmlEncodedCharacter(c));
                                } break;
                            }
                        }
                    }
                }
            }
        }
	}
        back = new String(sb);
        return back;
    }


    /**
     * Convert all backslashes to slashes in a string.
     * @param s Original string.
     * @return  Processed string.
     */
    public String backslashesToSlashes(String s) {
        String back = null;
	int i;
	char c;
	StringBuilder sb = new StringBuilder();
	if(s != null) {
	    for(i = 0; i < s.length(); i++) {
	        c = s.charAt(i);
		if(c == '\\') {
		    sb.append('/');
		} else {
		    sb.append(c);
		}
	    }
	}
	back = new String(sb);
	return back;
    }


    /**
     * Create a normalized date (german date format).
     * @param d Day.
     * @param m Month.
     * @param y Year.
     * @return  Date as string.
     */
    static
    private String createNormalizedDate(String d, String m, String y) {
        String back = null;
	int id = 0;
	int im = 0;
	int iy = 0;
	try {
	    id = Integer.parseInt(d);
	    im = Integer.parseInt(m);
	    iy = Integer.parseInt(y);
	    if((id >= 1) && (id <=31)) {
	        if((im >= 1) && (im <= 12)) {
		    if(iy < 100) {
		        iy += 2000;
		    }
		    String sd = (new Integer(id)).toString();
		    String sm = (new Integer(im)).toString();
		    String sy = (new Integer(iy)).toString();
		    StringBuilder sb = new StringBuilder();
		    if(sd.length() < 2) {
		        sb.append('0');
		    }
		    sb.append(sd);
		    sb.append('.');
		    if(sm.length() < 2) {
		        sb.append('0');
		    }
		    sb.append(sm);
		    sb.append('.');
		    sb.append(sy);
		    back = new String(sb);
		}
	    }
	}
	catch(Exception e) {
	}
	return back;
    }


    /**
     * Convert string to a normalized date (german date format).
     * @param s Source string.
     * @return  Date as string.
     */
    static
    public String getNormalizedDate(String s) {
        String back = null;
	String d;
	String m;
	String y;
	String s1;
	int i;
	if(s != null) {
	    i = s.indexOf('.');
	    if(i >= 0) {
	        d = s.substring(0, i);
		s1 = s.substring(i + 1);
		i = s1.indexOf('.');
		if(i >= 0) {
		    m = s1.substring(0, i);
		    y = s1.substring(i + 1);
		    back = createNormalizedDate(d, m, y);
		}
	    } else {
	        i = s.indexOf('-');
		if(i >= 0) {
		    y = s.substring(0, i);
		    s1 = s.substring(i + 1);
		    i = s1.indexOf('-');
		    if(i >= 0) {
		        m = s1.substring(0, i);
			d = s1.substring(i + 1);
			back = createNormalizedDate(d, m, y);
		    }
		} else {
		    i = s.indexOf('/');
		    if(i >= 0) {
		        y = s.substring(0, i);
			s1 = s.substring(i + 1);
			i = s1.indexOf('/');
			if(i >= 0) {
			    m = s1.substring(0, i);
			    d = s1.substring(i + 1);
			    back = createNormalizedDate(d, m, y);
			}
		    }
		}
	    }
	}
	return back;
    }


    /**
     * Get the current date in normalized form as string (german date format).
     * @return  Date string.
     */
    public static String getCurrentDateGerman() {
        String back = stt[8];
        Calendar cal = new GregorianCalendar();
        int y = cal.get(Calendar.YEAR);
        int m = cal.get(Calendar.MONTH);
        int d = cal.get(Calendar.DAY_OF_MONTH);
        if(y < 100) { y = 1900 + y; }
        String sy = (new Integer(y)).toString();
        String sm = (new Integer(m+1)).toString();
        String sd = (new Integer(d)).toString();
        StringBuilder sb = new StringBuilder();
        if(sd.length() < 2) { sb.append('0'); }
        sb.append(sd);
        sb.append('.');
        if(sm.length() < 2) { sb.append('0'); }
        sb.append(sm);
        sb.append('.');
        sb.append(sy);
        back = new String(sb);
        return back;
    }


    /* ********** GUI positioning ********** */


    /**
     * Choose value for one co-ordinate.
     *
     * @param sp
     * Start value of parent component.
     *
     * @param wp
     * Width value of parent component.
     *
     * @param wc
     * Width value of child component.
     *
     * @param wd
     * Width value of screen.
     *
     * @return
     * A start value for the child component.
     */
    private static int cp(int sp, int wp, int wc, int wd)
    {
        int back = 0;
        if((sp + wc) <= wd) {
            back = sp;
        } else {
            if((sp + wp) >= wc) {
                back = sp + wp - wc;
                // back = sp + wp - wc + 1;
            } else {
                back = sp + ((wp - wc) / 2);
            }
        }
        return back;
    }

    /**
     * Correct value for one co-ordinate
     *
     * @param x
     * The original (uncorrected) value.
     *
     * @param cw
     * Component width in this direction.
     *
     * @param dw
     * The screen width in this direction.
     *
     * @return
     * The corrected value.
     */
    private static int cor(int x, int cw, int dw)
    {
        int back = x;
        if((x + cw) > dw) {
            back = dw - cw;
        }
        if(back < 0) {
            back = 0;
        }
        return back;
    }

  /**
   * Choose position for component c relative to parent component p.
   *
   * @param w
   * Screen width.
   *
   * @param h
   * Screen height.
   *
   * @param p
   * Parent component.
   *
   * @param c
   * New child component to be placed.
   *
   * @param mod
   * Flag to indicate whether or not c is a modal component.
   */
    public static void choosePosition(
        int w, int h, Component p, Component c, boolean mod
    )
    {
        int x = -1;
        int y = -1;
        Rectangle pr = new Rectangle();
        Rectangle cr = new Rectangle();
	if((p != null) && (c != null)) {
            p.getBounds(pr);
            c.getBounds(cr);
            boolean b = false;
            if(!mod) {
                // right
                if((pr.x + pr.width + cr.width) < w) {
                    x = pr.x + pr.width;
                    y = cp(pr.y, pr.height, cr.height, h);
                    b = true;
                }
                // left
                if(!b) {
                    if(cr.width < pr.x) {
                        x = pr.x - cr.width;
                        y = cp(pr.y, pr.height, cr.height, h);
                        b = true;
                    }
                }
                // bottom
                if(!b) {
                    if((pr.y + pr.height + cr.height) < h) {
                        y = pr.y + pr.height;
                        x = cp(pr.x, pr.width, cr.width, w);
                        b = true;
                    }
                }
                // top
                if(!b) {
                    if(cr.height < pr.y) {
                        y = pr.y - cr.height;
                        x = cp(pr.x, pr.width, cr.width, w);
                        b = true;
                    }
                }
            }
            if(!b) {
                // above
                x = pr.x + ((pr.width - cr.width) / 2);
                y = pr.y + ((pr.height - cr.height) / 2);
            }
            x = cor(x, cr.width, w);
            y = cor(y, cr.height, h);
            c.setLocation(x, y);
	}
    }

    /**
     * Choose modal position for component c relative to parent component p.
     *
     * @param w
     * Screen width.
     *
     * @param h
     * Screen height.
     *
     * @param p
     * Parent component.
     *
     * @param c
     * New modal child component to be placed.
     *
     */
    public static void chooseModalPosition(
        int w, int h, Component p, Component c
    )
    {
        choosePosition(w, h, p, c, true);
    }

    /**
     * Choose normal (non-modal) position for component c relative to
     * parent component p.
     *
     * @param w
     * Screen width.
     *
     * @param h
     * Screen height.
     *
     * @param p
     * Parent component.
     *
     * @param c
     * New non-modal child component to be placed.
     */
    public static void chooseNonModalPosition(
        int w, int h, Component p, Component c
    )
    {
        choosePosition(w, h, p, c, false);
    }


    /**
     * Correct position after resizing.
     *
     * @param w
     * Screen width.
     *
     * @param h
     * Screen height.
     * 
     * @param c
     * The component for which the position must be corrected.
     */
    /*
    public static void correctPosition(int w, int h, Component c) {
        Rectangle cr = new Rectangle();
	if(c != null) {
            c.getBounds(cr);
            int x = cor(cr.x, cr.width, w);
            int y = cor(cr.y, cr.height, h);
            if((x != cr.x) || (y != cr.y)) {
                c.setLocation(x, y);
            }
	}
    } */

    
}
