/*
Copyright (c) 2010, Dirk Krause
All rights reserved.

Redistribution and use in source and binary forms,
with or without modification, are permitted provided
that the following conditions are met:

* Redistributions of source code must retain the above
  copyright notice, this list of conditions and the
  following disclaimer.
* Redistributions in binary form must reproduce the above 
  opyright notice, this list of conditions and the following
  disclaimer in the documentation and/or other materials
  provided with the distribution.
* Neither the name of the Dirk Krause nor the names of
  contributors may be used to endorse or promote
  products derived from this software without specific
  prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED.
IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
*/

package traceccgui;

import java.io.File;
import java.util.ArrayList;
import java.util.TreeMap;

import dirk_krause.tools.*;

/**
 *
 * @author krause
 */
public class TraceccController {

    /** Application controller. */
    protected DkTool dkt = null;

    /** Boolean settings. */
    protected boolean[] myBool = {
        false,      //  0: make mode
        true,       //  1: line number information
        false,      //  2: C++ style comments
        false,      //  3: produce debug output
        false,      //  4: debug output to stdout
        false,      //  5: prepare debug output for IDE
        false,      //  6: timestamp in debug output
        false,      //  7: trace keyword in debug output
        true,       //  8: .ctr .c
        true,       //  9: .cpt .cpp
        true,       // 10: .mtr .m
        true,       // 11: .jtr .java
        true,       // 12: .jl .java
    };

    /** Integer settings. */
    protected int[] myInt = {
        0,          //  0: IDE (gcc, MSVC, Workshop, tasm)
        75,         //  1: box width
    };

    /** String settings. */
    protected String[] myStr = {
        ".",        //  0: Current directory
    };

    /** Names to save boolean settings. */
    public static final String[] myPkBool = {
        "make",
        "output.lineno",
        "output.cplusplus",
        "debug.use",
        "debug.stdout",
        "debug.ide.use",
        "debug.timestamp",
        "debug.trace",
        "suffix.ctr",
        "suffix.cpt",
        "suffix.mtr",
        "suffix.jtr",
        "suffix.jl",
    };

    /** Names to save integer settings. */
    public static final String[] myPkInt = {
        "debug.ide.style",
        "output.comment-box",
    };

    /** Names to save string settings. */
    public static final String[] myPkStr = {
        "directory",
    };

    /** Suffixes the Tracecc GUI can handle. */
    public static final String[] suffixesToHandle = {
        ".ctr", ".cpt", ".mtr", ".jtr", ".jl"
    };

    public static final String[] outputSuffixes = {
        ".c", ".cpp", ".m", ".java", ".java"
    };

    public static final String[] kw = {
        "tracecc",                      //  0
        "jlayout",                      //  1
        "-r",                           //  2
        "--/ui/lang/env=off",           //  3
        "--/ui/lang=",                  //  4
        ".UTF-8",                       //  5
        "--/log/stdout/level=info",     //  6
        "--/log/stderr/level=none",     //  7
        "--/log/file/level=none",       //  8
        "--/log/stdout/time=yes",       //  9
        "--/log/stdout/split=yes",      // 10
        "--/log/file/keep=none",        // 11
        "--/log/stdout/codepage=-",     // 12
	"-l",                           // 13
	"-p",                           // 14
	"-d",                           // 15
	"-s",                           // 16
	"-i",                           // 17
	"gcc",                          // 18
	"msvc"  ,                       // 19
	"workshop",                     // 20
	"tasm",                         // 21
	"-t",                           // 22
	"-k",                           // 23
	"-b",                           // 24
    };

    /**
     * Constructor.
     * @param new_dkt New application controller.
     */
    public TraceccController(DkTool new_dkt)  {
        dkt = new_dkt;
        dkt.getMultiplePreferenceBools(myPkBool, myBool);
        dkt.getMultiplePreferenceInts(myPkInt, myInt);
        dkt.getMultiplePreferenceStrings(myPkStr, myStr);
    }

    /**
     * Save settings before the application exits.
     */
    public void saveSettings() {
        dkt.setMultiplePreferenceBools(myPkBool, myBool);
        dkt.setMultiplePreferenceInts(myPkInt, myInt);
        dkt.setMultiplePreferenceStrings(myPkStr, myStr);
    }

    /**
     * Set directory to process.
     * @param dn    Directory name.
     */
    public void setDirectory(String dn) {
        synchronized(this) {
            myStr[0] = dn;
        }
    }

    /**
     * Retrieve a boolean setting.
     * @param i Index of setting to retrieve.
     * @return  Setting value.
     */
    public boolean getBool(int i) {
        boolean back = false;
        synchronized(this) {
            if((i >= 0) && (i < myBool.length)) {
                back = myBool[i];
            }
        }
        return back;
    }

    /**
     * Retrieve string setting.
     * @param i Index of setting to retrieve.
     * @return  Setting value.
     */
    public String getStr(int i) {
        String back = null;
        synchronized(this) {
            if((i >= 0) && (i < myStr.length)) {
                back = myStr[i];
            }
        }
        return back;
    }

    /**
     * Set a string setting.
     * @param i Index.
     * @param v New value.
     */
    public void setStr(int i, String v) {
        synchronized(this) {
            if((i >= 0) && (i < myStr.length)) {
                myStr[i] = v;
            }
        }
    }

    /**
     * Retrieve all settings bundled.
     * @param b Boolean results array (may be null).
     * @param i Integer results array (may be null).
     * @param s String results array (may be null).
     */
    protected void getData(boolean[] b, int[] i, String[] s) {
        int j;
        int max;
        synchronized(this) {
            if(b != null) {
                max = myBool.length;
                if(max > b.length) { max = b.length; }
                for(j = 0; j < max; j++) {
                    b[j] = myBool[j];
                }
            }
            if(i != null) {
                max = myInt.length;
                if(max > i.length) { max = i.length; }
                for(j = 0; j < max; j++) {
                    i[j] = myInt[j];
                }
            }
            if(s != null) {
                max = myStr.length;
                if(max > s.length) { max = s.length; }
                for(j = 0; j < max; j++) {
                    s[j] = myStr[j];
                }
            }
        }
    }

    /**
     * Set bundled settings.
     * @param b Boolean settings (may be null).
     * @param i Integer settings (may be null).
     * @param s String settings (may be null).
     */
    protected void putData(boolean[] b, int[] i, String[] s) {
        int j;
        int max;
        synchronized(this) {
            if(b != null) {
                max = myBool.length;
                if(max > b.length) { max = b.length; }
                for(j = 0; j < max; j++) {
                    myBool[j] = b[j];
                }
            }
            if(i != null) {
                max = myInt.length;
                if(max > i.length) { max = i.length; }
                for(j = 0; j < max; j++) {
                    myInt[j] = i[j];
                }
            }
            if(s != null) {
                max = myStr.length;
                if(max > s.length) { max = s.length; }
                for(j = 0; j < max; j++) {
                    myStr[j] = s[j];
                }
            }
        }
    }


    /**
     * Create command set which is run by the main window.
     * @return  Command set (may be null).
     */
    public TraceccCommandSet createListOfCommands() {
        // Function result.
        TraceccCommandSet back = null;
        // Private copy of the myBool
        boolean[] b = new boolean[myBool.length];
        // Private copy of myInt
        int[] i = new int[myInt.length];
        // Private copy of MyStr
        String[] s = new String[myStr.length];
        // Map filename - conversion command line
        TreeMap<String,String[]> tm = new TreeMap<String,String[]>();
        // Flag: Can handle this file.
        boolean can_handle;
        // Flag: Must handle this file.
        boolean must_handle;
        // Suffix of current file.
        String suf;
        // Flag: Backslash if file name part separator.
        boolean useBackslash = false;
        // Flag: File names are case sensitive.
        boolean caseSensitive = true;
        if(dkt.getFileSeparator().equals("\\")) {
            caseSensitive = false; useBackslash = true;
        }
        // Input file name.
        String ifn;
        // Output file name.
        String ofn;

        getData(b, i, s);
        if(s[0] != null) {
            File curDir = new File(s[0]);
            if(curDir.exists()) {
                if(curDir.isDirectory()) {
                    String[] allFiles = curDir.list();
                    for(String filename: allFiles) {
                        can_handle = false; must_handle = false;
                        suf = dkt.getFilenameSuffix(filename);
                        if(useBackslash) {
                            ifn = s[0] + "\\" + filename;
                        } else {
                            ifn = s[0] + "/" + filename;
                        }
                        File fifn = new File(ifn);
                        if(fifn.exists() && fifn.isFile()) {
                            if(suf != null) {
                                int sft = DkTool.getStringArrayIndex(
                                    suffixesToHandle, suf, caseSensitive
                                );
                                if(sft >= 0) {
                                    ofn = dkt.replaceSuffix(
                                        ifn, outputSuffixes[sft]
                                    );
                                    if(b[8 + sft]) {
                                        can_handle = true; must_handle = true;
                                        if(b[0]) {
                                            File fofn = new File(ofn);
                                            if(fofn.exists() && fofn.isFile())
                                            {
                                                long lmo = fofn.lastModified();
                                                long lms = fifn.lastModified();
                                                if(lmo > lms) {
                                                    must_handle = false;
                                                }
                                            }
                                        }
                                    }
                                    if(must_handle) {
                                        ArrayList<String> argv =
                                        new ArrayList<String>();
                                        argv.add(kw[(sft == 4) ? 1 : 0]);
                                        String lang = dkt.getLanguage();
                                        if(lang.indexOf('.') < 0) {
                                            lang = lang + kw[5];
                                        }
                                        argv.add(kw[3]);
                                        argv.add(kw[4] + lang);
                                        for(int x = 6; x < 13; x++) {
                                            argv.add(kw[x]);
                                        }
                                        if(sft < 4) {
					    if(b[1]) {
					        argv.add(kw[13]);
					    }
					    if(b[2]) {
					        argv.add(kw[14]);
					    }
					    argv.add(kw[24]);
					    argv.add(
                                                (new Integer(i[1])).toString()
                                            );
					    if(b[3]) {
					        argv.add(kw[15]);
						if(b[4]) {
						    argv.add(kw[16]);
						}
						if(b[6]) {
						    argv.add(kw[22]);
						}
						if(b[7]) {
						    argv.add(kw[23]);
						}
						if(b[5]) {
						    if((i[0] >= 0)
                                                       && (i[0] <= 33))
                                                    {
						        argv.add(kw[17]);
							argv.add(kw[18 + i[0]]);
						    }
						}
					    }
                                        }
                                        argv.add(ifn);
                                        argv.add(ofn);
                                        Object[] oa = argv.toArray();
                                        int oal = oa.length;
                                        String[] cmdl = new String[oal];
                                        int cmdi;
                                        for(cmdi = 0; cmdi < oal; cmdi++) {
                                            cmdl[cmdi] =
                                            new String((String)(oa[cmdi]));
                                        }
                                        tm.put(ifn, cmdl);
                                    }
                                }
                            }
                        }
                    }
                }
            }
            back = new TraceccCommandSet(s[0], tm);
        }
        return back;
    }

}
